# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

import textwrap
import pytest
import numpy as np
import networkx as nx
import vermouth.forcefield
import vermouth.molecule
import polyply.src.meta_molecule
import polyply.src.map_to_molecule
import polyply.src.polyply_parser
from polyply.src.meta_molecule import (MetaMolecule, Monomer)
from vermouth.molecule import Interaction

@pytest.mark.parametrize('lines, monomers, bonds, edges, nnodes, from_itp', (
    # single block
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    2  SN1a    1   PEO   CO2  1   0.000  45
    3  SN1a    1   PEO   CO3  1   0.000  45
    4  SN1a    1   PEO   CO4  1   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 8000
    3  4   1   0.37 9000
    """,
    ["PEO", "PEO"],
    [Interaction(atoms=(0, 1), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(1, 2), parameters=['1', '0.37', '8000'], meta={}),
     Interaction(atoms=(2, 3), parameters=['1', '0.37', '9000'], meta={}),
     Interaction(atoms=(4, 5), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(5, 6), parameters=['1', '0.37', '8000'], meta={}),
     Interaction(atoms=(6, 7), parameters=['1', '0.37', '9000'], meta={})],
    [(0, 1), (1, 2), (2, 3), (4, 5), (5, 6), (6, 7)],
    8,
    {}),
    # linear multiblock last
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         1
    ;
    [ atoms ]
    1  SN1a    1   R1   C1  1   0.000  45
    2  SN1a    1   R1   C2  1   0.000  45
    3  SC1     2   R2   C1  2   0.000  45
    4  SC1     2   R2   C2  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 7000
    3  4   1   0.37 7000
    """,
    ["PEO", "MIX"],
    [Interaction(atoms=(1, 2), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(2, 3), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(3, 4), parameters=['1', '0.37', '7000'], meta={})],
    [(1, 2), (2, 3), (3, 4)],
    5,
    {}),
    # linear multi-block second
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         1
    ;
    [ atoms ]
    1  SN1a    1   R1   C1  1   0.000  45
    2  SN1a    1   R1   C2  1   0.000  45
    3  SC1     2   R2   C1  2   0.000  45
    4  SC1     2   R2   C2  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 7000
    3  4   1   0.37 7000
    """,
    ["MIX", "PEO"],
    [Interaction(atoms=(0, 1), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(1, 2), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(2, 3), parameters=['1', '0.37', '7000'], meta={})],
    [(0, 1), (1, 2), (2, 3)],
    5,
    {}),
    # linear branched
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         1
    ;
    [ atoms ]
    1  SN1a    1   R1   C1  1   0.000  45
    2  SC1     1   R1   C1  2   0.000  45
    3  SN1a    2   R2   C1  1   0.000  45
    4  SC1     2   R2   C1  2   0.000  45
    5  SN1a    3   R3   C1  1   0.000  45
    6  SC1     3   R3   C1  2   0.000  45
    7  SN1a    4   R2   C1  1   0.000  45
    8  SC1     4   R2   C1  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.44 7000
    1  3   1   0.37 7000
    3  4   1   0.44 7000
    3  5   1   0.37 7000
    3  7   1   0.37 7000
    5  6   1   0.44 7000
    7  8   1   0.44 7000
    """,
    ["PEO", "MIX"],
    [Interaction(atoms=(1, 2), parameters=['1', '0.44', '7000'], meta={}),
     Interaction(atoms=(1, 3), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(3, 4), parameters=['1', '0.44', '7000'], meta={}),
     Interaction(atoms=(3, 5), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(3, 7), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(5, 6), parameters=['1', '0.44', '7000'], meta={}),
     Interaction(atoms=(7, 8), parameters=['1', '0.44', '7000'], meta={})],
    [(1, 2), (1, 3), (3, 4), (3, 5), (3, 7), (5, 6), (7, 8)],
    9,
    {}),
    # test multiblock from-itp
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         1
    ;
    [ atoms ]
    1  SN1a    1   R1   C1  1   0.000  45
    2  SN1a    1   R1   C2  1   0.000  45
    3  SC1     2   R2   C1  2   0.000  45
    4  SC1     2   R2   C2  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 7000
    3  4   1   0.37 7000
    """,
    ["R1", "R2", "PEO"],
    [Interaction(atoms=(0, 1), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(1, 2), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(2, 3), parameters=['1', '0.37', '7000'], meta={})],
    [(0, 1), (1, 2), (2, 3)],
    5,
    {0: "MIX", 1: "MIX"}),
    # test multiblock from-itp reverse
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         1
    ;
    [ atoms ]
    1  SN1a    1   R1   C1  1   0.000  45
    2  SN1a    1   R1   C2  1   0.000  45
    3  SC1     2   R2   C1  2   0.000  45
    4  SC1     2   R2   C2  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 7000
    3  4   1   0.37 7000
    """,
    ["PEO", "R1", "R2"],
    [Interaction(atoms=(1, 2), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(2, 3), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(3, 4), parameters=['1', '0.37', '7000'], meta={})],
    [(1, 2), (2, 3), (3, 4)],
    5,
    {1: "MIX", 2: "MIX"}),
    # test two multiblock from-itp
    ("""
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         1
    ;
    [ atoms ]
    1  SN1a    1   R1   C1  1   0.000  45
    2  SN1a    1   R1   C2  1   0.000  45
    3  SC1     2   R2   C1  2   0.000  45
    4  SC1     2   R2   C2  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 7000
    3  4   1   0.37 8000
    """,
    ["R1", "R2", "PEO", "R1", "R2"],
    [Interaction(atoms=(0, 1), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(1, 2), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(2, 3), parameters=['1', '0.37', '8000'], meta={}),
     Interaction(atoms=(5, 6), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(6, 7), parameters=['1', '0.37', '7000'], meta={}),
     Interaction(atoms=(7, 8), parameters=['1', '0.37', '8000'], meta={})],
    [(0, 1), (1, 2), (2, 3), (5, 6), (6, 7), (7, 8)],
    9,
    {0: "MIX", 1: "MIX", 3: "MIX", 4:"MIX"})
    ))
def test_multiresidue_block(lines, monomers, bonds, edges, nnodes, from_itp):
    """
    Test multiresidue blocks are correctly identified and parameters from
    the blocks added at the correct place in the fine-grained molecule.
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    polyply.src.polyply_parser.read_polyply(lines, ff)
    # build the meta-molecule
    meta_mol = MetaMolecule(name="test", force_field=ff)
    meta_mol.add_monomer(0, monomers[0], [])
    for node, monomer in enumerate(monomers[1:]):
        meta_mol.add_monomer(node+1, monomer, [(node, node+1)])
    nx.set_node_attributes(meta_mol, from_itp, "from_itp")
    # map to molecule
    new_meta_mol = polyply.src.map_to_molecule.MapToMolecule(ff).run_molecule(meta_mol)
    # check that the disconnected molecule is properly done
    for node in new_meta_mol.nodes():
        print(new_meta_mol.nodes[node]['graph'].nodes)
    assert len(new_meta_mol.molecule.nodes) == nnodes
    assert list(new_meta_mol.molecule.edges) == edges
    assert new_meta_mol.molecule.interactions['bonds'] == bonds

def test_multi_excl_block():
    lines = """
    [ moleculetype ]
    ; name nexcl.
    PEO         1
    ;
    [ atoms ]
    1  SN1a    1   PEO   CO1  1   0.000  45
    [ moleculetype ]
    ; name nexcl.
    MIX         2
    ;
    [ atoms ]
    1  SN1a    1   MIX  C1  1   0.000  45
    2  SN1a    1   MIX  C2  1   0.000  45
    3  SC1     1   MIX  C1  2   0.000  45
    4  SC1     1   MIX  C2  2   0.000  45
    [ bonds ]
    ; back bone bonds
    1  2   1   0.37 7000
    2  3   1   0.37 7000
    3  4   1   0.37 7000
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    polyply.src.polyply_parser.read_polyply(lines, ff)
    meta_mol = MetaMolecule(name="test", force_field=ff)
    meta_mol.add_monomer(0,"PEO",[])
    meta_mol.add_monomer(1,"MIX",[(1,0)])

    new_meta_mol = polyply.src.map_to_molecule.MapToMolecule(ff).run_molecule(meta_mol)
    assert nx.get_node_attributes(new_meta_mol.molecule, "exclude") == {0: 1, 1: 2, 2: 2, 3: 2, 4: 2}



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
from polyply.src.processor import Processor

def tag_exclusions(node_to_block, force_field):
    """
    Given block names matching nodes in the meta_molecule
    graph check if the corresponding blocks in `force_field`
    all have the same number of default exclusions. If not
    find the minimum number of exclusions and tag all
    nodes with the original exclusion number. Then
    change the exclusion number to the lowest value.

    Note the tag is picked up by apply links where
    the excluions are generated.
    """
    excls = {}
    for node in node_to_block:
        block = force_field.blocks[node_to_block[node]]
        excls[node] = block.nrexcl

    if len(set(excls.values())) > 1:
        min_excl = min(list(excls.values()))
        for node, excl in excls.items():
            block = force_field.blocks[node_to_block[node]]
            nx.set_node_attributes(block, excl, "exclude")
            block.nrexcl = min_excl

def _correspondence_to_residue(meta_molecule,
                               molecule,
                               correspondence,
                               res_node):
    """
    Given a `meta_molecule` and the underlying higher resolution
    `molecule` as well as a correspondence dict, describing how
    a single node (res_node) in meta_molecule corresponds to a
    fragment in molecule make a graph of that residue and propagate
    all meta_molecule node attributes to that graph.

    Parameters
    ----------
    meta_molecule: polyply.src.meta_molecule.MetaMolecule
        The meta molecule to process.
    molecule: vermouth.molecule.Molecule
    correspondance: list
    res_node: abc.hashable
        the node in meta_molecule
    """
    resid = meta_molecule.nodes[res_node]["resid"]
    residue = nx.Graph()
    for mol_node in correspondence.values():
        data = molecule.nodes[mol_node]
        if data["resid"] == resid:
            residue.add_node(mol_node, **data)
            for attribute, value in meta_molecule.nodes[res_node].items():
                # graph and seqID are specific attributes set by make residue
                # graph or the gen_seq tool, which we don't want to propagate.
                if attribute in ["graph", "seqID"]:
                    continue
                residue.nodes[mol_node][attribute] = value

    return residue

class MapToMolecule(Processor):
    """
    This processor takes a :class:`MetaMolecule` and generates a
    :class:`vermouth.molecule.Molecule`, which consists at this stage
    of disconnected blocks. These blocks can be connected using the
    :class:`ApplyLinks` processor.
    """
    def __init__(self, force_field):
        self.node_to_block = {}
        self.node_to_fragment = {}
        self.fragments = []
        self.multiblock_correspondence = []
        self.added_fragments = []
        self.added_fragment_nodes = []
        self.force_field = force_field

    def match_nodes_to_blocks(self, meta_molecule):
        """
        This function matches the nodes in the meta_molecule
        to the blocks in the force-field. It does the essential
        bookkeeping for three cases and populates the node_to_block,
        and node_to_fragment dicts as well as the fragments attribute.
        It distinguishes three cases:

        1) the node corresponds to a single residue block; here
           node_to_block entry is simply the resname of the block
        2) the node has the from_itp attribute; in this case
           the node is part of a multiresidue block in the FF,
           all nodes corresponding to that block form a fragment.
           All fragments are added to the fragments attribute, and
           the nodes in those fragments all have the entry node_to_block
           set to the block. In addition it is recorded to which fragment
           specifically the node belongs in the node_to_fragment dict.
        3) the node corresponds to a multiresidue block; but unlike in
           case to it represents multiple residues. ....

         Parameters
        ----------
        meta_molecule: polyply.src.meta_molecule.MetaMolecule
            The meta molecule to process.

        """
        regular_graph = nx.Graph()
        restart_graph = nx.Graph()
        restart_attr = nx.get_node_attributes(meta_molecule, "from_itp")

        # this breaks down when to proteins are directly linked
        # because they would appear as one connected component
        # and not two seperate components referring to two molecules
        # but that is an edge-case we can worry about later
        for idx, jdx in nx.dfs_edges(meta_molecule):
            # the two nodes are restart nodes
            if idx in restart_attr and jdx in restart_attr:
                restart_graph.add_edge(idx, jdx)
            else:
                regular_graph.add_edge(idx, jdx)

        # regular nodes have to match a block in the force-field by resname
        for node in regular_graph.nodes:
            self.node_to_block[node] = meta_molecule.nodes[node]["resname"]

        # fragment nodes match parts of blocks, which describe molecules
        # with more than one residue
        for fragment in nx.connected_components(restart_graph):
            block_name = restart_attr[list(fragment)[0]]
            if all([restart_attr[node] == block_name  for node in fragment]):
                self.fragments.append(fragment)
                block = self.force_field.blocks[block_name]
                for node in fragment:
                    self.node_to_block[node] = block_name
                    self.node_to_fragment[node] = len(self.fragments) - 1
            else:
                raise IOError

    def add_blocks(self, meta_molecule):
        """
        Add disconnected blocks to :class:`vermouth.molecule.Moleclue`
        and set the graph attribute to meta_molecule matching the node
        with the underlying fragment it represents at higher resolution.
        Note that this function also takes care to properly add multi-
        residue blocks (i.e. from an existing itp-file).

        Parameters
        ----------
        meta_molecule: polyply.src.meta_molecule.MetaMolecule
            The meta molecule to process.

        Returns
        -------
        vermouth.molecule.Molecule
            The disconnected fine-grained molecule.
        """
        # get a defined order for looping over the resiude graph
        node_keys = list(meta_molecule.nodes())
        resid_dict = nx.get_node_attributes(meta_molecule, "resid")
        resids = [resid_dict[node] for node in node_keys]
        node_keys = [x for _, x in sorted(zip(resids, node_keys))]
        # get the first node and convert it to molecule
        start_node = node_keys[0]
        new_mol = self.force_field.blocks[self.node_to_block[start_node]].to_molecule()

        # in this case the node belongs to a fragment for which there is a
        # multiresidue block
        if "from_itp" in meta_molecule.nodes[start_node]:
            # add all nodes of that fragment to added_fragment nodes
            fragment_nodes = list(self.fragments[self.node_to_fragment[start_node]])
            self.added_fragment_nodes += fragment_nodes

            # extract the nodes of this paticular residue and store a
            # dummy correspndance
            correspondence = {node:node for node in new_mol.nodes}
            self.multiblock_correspondence.append({node:node for node in new_mol.nodes})
            residue = _correspondence_to_residue(meta_molecule,
                                                 new_mol,
                                                 correspondence,
                                                 start_node)
            # add residue to meta_molecule node
            meta_molecule.nodes[start_node]["graph"] = residue
        else:
            # we store the block together with the residue node
            meta_molecule.nodes[start_node]["graph"] = new_mol.copy()

        # now we loop over the rest of the nodes
        for node in node_keys[1:]:
            # in this case the node belongs to a fragment which has been added
            # we only extract the residue belonging to this paticular node
            if node in self.added_fragment_nodes:
                fragment_id = self.node_to_fragment[node]
                correspondence = self.multiblock_correspondence[fragment_id]
            # in this case we have to add the node from the block definitions
            else:
                block = self.force_field.blocks[self.node_to_block[node]]
                correspondence = new_mol.merge_molecule(block)

            # make the residue from the correspondence
            residue = _correspondence_to_residue(meta_molecule,
                                                 new_mol,
                                                 correspondence,
                                                 node)
            # add residue to node
            meta_molecule.nodes[node]["graph"] = residue

            # in this case we just added a new multiblock residue so we store
            # the correspondence as well as keep track of the nodes that are
            # part of that fragment
            if "from_itp" in meta_molecule.nodes[node] and node not in self.added_fragments:
                fragment_nodes = list(self.fragments[self.node_to_fragment[node]])
                self.added_fragment_nodes += fragment_nodes
                self.multiblock_correspondence.append(correspondence)

        return new_mol

    def run_molecule(self, meta_molecule):
        """
        Take a meta_molecule and generated a disconnected graph
        of the higher resolution molecule by matching the resname
        attribute to blocks in the force-field. This function
        also takes care to correcly add parameters from an itp
        file to fine-grained molecule. It also sets the 'graph'
        attribute, which is the higher-resolution fragment that
        the meta_molecule node represents.

        Parameters
        ----------
        molecule: polyply.src.meta_molecule.MetaMolecule
             The meta molecule to process.

        Returns
        -------
        molecule: polyply.src.meta_molecule.MetaMolecule
            The meta molecule with attribute molecule that is the
            fine grained molecule.
        """
        # in a first step we match the residue names to blocks
        # in the force-field. Residue names can also be part
        # of a larger fragment stored as a block or refer to
        # a block which consists of multiple residues. This
        # gets entangled here
        self.match_nodes_to_blocks(meta_molecule)
        # next we check if all exclusions are the same and if
        # not we adjust it such that the lowest exclusion number
        # is used. ApplyLinks then generates those appropiately
        tag_exclusions(self.node_to_block, self.force_field)
        # now we add the blocks generating a new molecule
        new_molecule = self.add_blocks(meta_molecule)
        meta_molecule.molecule = new_molecule
        return meta_molecule



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test the nonbond engine computes forces and such accurately
"""
import pytest
import textwrap
import numpy as np
from vermouth.forcefield import ForceField
import polyply
from polyply.src.topology import Topology
from polyply.src.top_parser import read_topology
from polyply.src.nonbond_engine import NonBondEngine

@pytest.fixture
def topology():
    top_lines ="""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 72.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0  1   GLY    BB   1 0.00     45
    2    N0  1   GLY    SC1  1 0.00     45
    3    N0  1   GLY    SC2  1 0.00     45
    4    N0  2   GLU    BB   2 0.00     45
    5    N0  2   GLU    SC1  2 0.00     45
    6    N0  2   GLU    SC2  2 0.00     45
    [ bonds ]
    1    2    1  0.47 2000
    2    3    1  0.47 2000
    1    4    1  0.47 2000
    4    5    1  0.47 2000
    4    6    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0  1   ASP    BB   1 0.00     45
    2    N0  1   ASP    SC1  1 0.00     45
    3    N0  1   ASP    SC2  1 0.00     45
    [ bonds ]
    1    2    1  0.47 2000
    2    3    1  0.47 2000
    [ system ]
    test system
    [ molecules ]
    testA 2
    testB 1
    """
    lines = textwrap.dedent(top_lines)
    lines = lines.splitlines()
    force_field = ForceField("test")
    topology = Topology(force_field)
    read_topology(lines=lines, topology=topology, cwdir="./")
    topology.preprocess()
    topology.volumes = {"GLY": 0.53, "GLU": 0.67, "ASP": 0.43}
    return topology

@pytest.mark.parametrize('n_pos', (
      0,
      3,
      5))
def  test_nb_engine_from_top_file(topology, n_pos):

     # we add positions to the topology skipping some
     # positins depending on how many are set in n_pos
     positions = np.random.random(n_pos*3).reshape(-1, 3)
     for mol_idx, mol in enumerate(topology.molecules):
         for node in mol.nodes:
             try:
                mol.nodes[node]["position"] = positions[mol_idx + node]
             except IndexError:
                continue

     # initiate the nb_engine
     nb_engine =  NonBondEngine.from_topology(topology.molecules,
                                              topology,
                                              box=np.array([10., 10., 10.]))

     # these are the expected interactions in the nb_engine
     nb_interactions = {
                        frozenset(["GLY", "GLY"]): 0.53,
                        frozenset(["GLY", "GLU"]): (0.53+0.67)/2.0,
                        frozenset(["GLY", "ASP"]): (0.43+0.53)/2.0,
                        frozenset(["ASP", "ASP"]): 0.43,
                        frozenset(["ASP", "GLU"]): (0.43+0.67)/2.0,
                        frozenset(["GLU", "GLU"]): 0.67,
                       }

     # make sure the cut_off is set dynamically and accurate
     assert nb_engine.cut_off == 2*0.67

     # check if all interactions are created properly
     for res_combo, interaction in nb_interactions.items():
          assert nb_engine.interaction_matrix[res_combo] == (nb_interactions[res_combo], 1.0)

     # make sure that all positions are in the global position matrix
     # and positions which are undefined have a infinity set
     for mol_idx, mol in enumerate(topology.molecules):
         for node in mol.nodes:
            if "position" in mol.nodes[node]:
                assert all(nb_engine.get_point(mol_idx, node) == mol.nodes[node]["position"])
            else:
                assert all(nb_engine.positions[mol_idx, node] == np.array([np.inf, np.inf, np.inf]))

def test_add_positions(topology):
     # we add 2 positions to the nb_matrix
     positions = np.random.random(2*3).reshape(-1, 3)
     for mol_idx, mol in enumerate(topology.molecules):
         for node in mol.nodes:
             try:
                mol.nodes[node]["position"] = positions[mol_idx + node]
             except IndexError:
                continue

     nb_engine =  NonBondEngine.from_topology(topology.molecules,
                                              topology,
                                              box=np.array([10., 10., 10.]))

     # now we check if the 3rd is added correctly
     nb_engine.add_positions(np.array([5.0, 5.0, 5.0]), mol_idx=2, node_key=0, start=False)
     assert all(nb_engine.get_point(2, 0) == np.array([5.0, 5.0, 5.0]))

def test_remove_positions(topology):
     # we add 5 positions to the nb_matrix
     positions = np.random.random(5*3).reshape(-1, 3)
     for mol_idx, mol in enumerate(topology.molecules):
         for node in mol.nodes:
             try:
                mol.nodes[node]["position"] = positions[mol_idx + node]
             except IndexError:
                continue

     nb_engine =  NonBondEngine.from_topology(topology.molecules,
                                              topology,
                                              box=np.array([10., 10., 10.]))

     # now we check if the 3rd is removed correctly
     nb_engine.remove_positions(mol_idx=2, node_keys=[0])
     assert all(nb_engine.get_point(2, 0) == np.array([np.inf, np.inf, np.inf]))

def test_update_positions_in_molecules(topology):
     # we add 15 random positions to all molecules
     positions = np.random.random(5*3).reshape(-1, 3)
     for mol_idx, mol in enumerate(topology.molecules):
         for node in mol.nodes:
             print(mol_idx, node)
             mol.nodes[node]["position"] = positions[mol_idx + node]

     nb_engine =  NonBondEngine.from_topology(topology.molecules,
                                              topology,
                                              box=np.array([10., 10., 10.]))
     # now we overwrite those positoins one
     nb_engine.positions[:, :] = np.ones((5, 3), dtype=float)[:, :]
     # then we update them in the molecules and check if they are actually updated
     nb_engine.update_positions_in_molecules(topology.molecules)
     for mol_idx, mol in enumerate(topology.molecules):
         for node in mol.nodes:
            assert all(mol.nodes[node]["position"] == np.array([1., 1., 1.]))

@pytest.mark.parametrize('mol_idx_a, mol_idx_b, node_a, node_b, expected',
                        ((0, 0, 0, 1, (0.53+0.67)/2.0),
                         (0, 2, 0, 0, (0.43+0.53)/2.0),
                         (0, 2, 1, 0, (0.43+0.67)/2.0)))
def test_get_interaction(topology, mol_idx_a, mol_idx_b, node_a, node_b, expected):
     # initiate the nb_engine
     nb_engine =  NonBondEngine.from_topology(topology.molecules,
                                              topology,
                                              box=np.array([10., 10., 10.]))

     # check get interactions works properly
     value = nb_engine.get_interaction(mol_idx_a=mol_idx_a,
                                       mol_idx_b=mol_idx_b,
                                       node_a=node_a,
                                       node_b=node_b)
     assert value == (expected, 1.0)

@pytest.mark.parametrize('dist, ref, expected',
                        # LJ force is zero at minimum
			((2**(1/6.)*0.35,
                          np.array([0.0, 0.0, 2**1/6.*0.35]),
                          np.array([0.0, 0.0, 0.0])),
                        # LJ force is also zero at infinity
                         (100000.0,
                          np.array([0.0, 0.0, 100000.0]),
                          np.array([0.0, 0.0, 0.0])),
                        # pick one point inbetween
                         (0.42,
                          np.array([0.0, 0.0, 0.42]),
                          np.array([0.0, 0.0, 13.27016])),
                        # distance along different vector
                         (0.42,
                          np.array([0.0, 0.42, 0.0]),
                          np.array([0.0, 13.27016, 0.0])),
                        # truly 3d distance
                         (0.42,
                          np.array([0.2, 0.4, 0.2]),
                          np.array([ 6.31912, 12.63825,  6.31912]))
                        ))
def test_LJ_force(dist, ref, expected):
     point = np.array([0.0, 0.0, 0.0])
     params = (0.35, 2.1)
     value = polyply.src.nonbond_engine._lennard_jones_force(dist, point, ref, params)
     print(value)
     assert np.allclose(value, expected)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

import textwrap
import pytest
import math
import vermouth.forcefield
import vermouth.molecule
from vermouth.molecule import Interaction
import polyply.src.meta_molecule
from polyply import TEST_DATA
from polyply.src.topology import Topology

class TestTopology:

    @staticmethod
    def test_from_gmx_topfile():
        top = Topology.from_gmx_topfile(TEST_DATA+"/topology_test/system.top", "test")
        assert len(top.molecules) == 1

    @staticmethod
    def test_add_positions_from_gro():
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
        top.add_positions_from_file(TEST_DATA + "/topology_test/test.gro")
        for node in top.molecules[0].molecule.nodes:
            if node < 14:
                assert "position" in top.molecules[0].molecule.nodes[node].keys()

        for node in top.molecules[0].nodes:
            if node != 2:
                assert "position" in top.molecules[0].nodes[node].keys()
                assert top.molecules[0].nodes[node]["build"] == False
            else:
                assert top.molecules[0].nodes[node]["build"] == True

    @staticmethod
    def test_add_positions_from_pdb():
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/pdb.top", "test")
        top.add_positions_from_file(TEST_DATA + "/topology_test/test.pdb")
        for meta_mol in top.molecules:
            for node in meta_mol.molecule.nodes:
                assert "position" in meta_mol.molecule.nodes[node].keys()

        for meta_mol in top.molecules:
            for node in meta_mol.nodes:
                    assert "position" in meta_mol.nodes[node].keys()
                    assert meta_mol.nodes[node]["build"] == False

    @staticmethod
    def test_convert_to_vermouth_system():
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
        system = top.convert_to_vermouth_system()
        assert isinstance(system, vermouth.system.System)
        assert len(system.molecules) == 1

    @staticmethod
    @pytest.mark.parametrize('lines, outcome', (
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        """,
        {frozenset(["O", "O"]): {"f": 1,
                               "nb1": 2.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "C"]): {"f": 1,
                               "nb1": 1.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "O"]): {"f": 1,
                               "nb1": 0.0022106496,
                               "nb2": 9.9002500e-07}}
        ),
        ("""
        [ defaults ]
        1.0   3.0    yes  0.5     0.5
        [ atomtypes ]
        C   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        O   O   8      15.99940    -0.500       A    2.96000e-01  8.78640e-01 ; SIG
        """,
        {frozenset(["C", "C"]): {"f": 1,
                               "nb1": 3.75000e-01,
                               "nb2": 4.39320e-01},
         frozenset(["O", "O"]): {"f": 1,
                               "nb1": 2.96000e-01,
                               "nb2": 8.78640e-01},
         frozenset(["O", "C"]): {"f": 1,
                               "nb1": 0.3355,
                               "nb2": 0.6212923022217481}}
        ),
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        [ nonbond_params ]
        C    O    1     2.0     4.0
        """,
        {frozenset(["O", "O"]): {"f": 1,
                               "nb1": 2.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "C"]): {"f": 1,
                               "nb1": 1.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "O"]): {"f": 1,
                               "nb1": 2.0,
                               "nb2": 4.0}}
        )))
    def test_gen_pairs(lines, outcome):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        top.gen_pairs()
        for atom_pair in outcome:
            assert atom_pair in top.nonbond_params
            nb1 = top.nonbond_params[atom_pair]["nb1"]
            nb2 = top.nonbond_params[atom_pair]["nb2"]
            nb1_ref = outcome[atom_pair]["nb1"]
            nb2_ref = outcome[atom_pair]["nb2"]
            assert math.isclose(nb1, nb1_ref)
            assert math.isclose(nb2, nb2_ref)

    @staticmethod
    @pytest.mark.parametrize('lines, outcome', (
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        #define ga_1  100  250
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CH3   1 test C1 1   0.0 14.0
        2 CH2   1 test C2 2   0.0 12.0
        3 CH3   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 2 ga_1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        ["2", "100", "250"]
        ),
        # different location of define statement
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        #define ga_1  100  250
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CH3   1 test C1 1   0.0 14.0
        2 CH2   1 test C2 2   0.0 12.0
        3 CH3   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 2 ga_1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        ["2", "100", "250"]
        ),
        # two defines for one statement
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        #define ga_1  100
        #define gk_1  250
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CH3   1 test C1 1   0.0 14.0
        2 CH2   1 test C2 2   0.0 12.0
        3 CH3   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 2 ga_1 gk_1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        ["2", "100", "250"]
        )
	))
    def test_replace_defines(lines, outcome):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        top.replace_defines()
        assert top.molecules[0].molecule.interactions["angles"][0].parameters == outcome

    @staticmethod
    def test_convert_nonbond_to_sig_eps():
        """
        Simply test if the conversion from C6 C12 to simga epsilon
        is done properly.
        """

        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        top.nonbond_params = {frozenset(["EO", "EO"]):
                             {"nb1":6.44779031E-02 , "nb2": 4.07588234E-04}}
        top.convert_nonbond_to_sig_eps()
        assert math.isclose(top.nonbond_params[frozenset(["EO", "EO"])]["nb1"], 0.43)
        assert math.isclose(top.nonbond_params[frozenset(["EO", "EO"])]["nb2"], 3.4*0.75)

    @staticmethod
    @pytest.mark.parametrize('lines, outcome', (
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ bondtypes ]
        C       C       1       0.1335  502080.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 C   1 test C1 1   0.0 14.0
        2 C   1 test C2 2   0.0 12.0
        [ bonds ]
        1 2  1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={})]}
        ),
        # test three element define
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ angletypes ]
        CE1   CE1	CT2	5	123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CE1   1 test C1 1   0.0 14.0
        2 CE1   1 test C2 2   0.0 12.0
        3 CT2   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"angles": [Interaction(atoms=(0, 1, 2), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={})]}
        ),
        # test reverse match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ angletypes ]
        CE1    CE2	CT2	5	123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        [ angles ]
        1  2  3 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"angles": [Interaction(atoms=(0, 1, 2), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={})]}
        ),
        # test generic match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        X  CE2  CE1  X    5   123.50	401.664	0.0	0.0
        X  CT2  CE1  X    5   120	400	0.0	0.0
        X  QQQ  QQQ  X    5   150	400	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        2  3  4  5 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={}),
                       Interaction(atoms=(1, 2, 3, 4), parameters=["5", "120", "400", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test generic match plus defined match on same pattern
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        X  CE2  CE1  X    5   123.50	401.664	0.0	0.0
        X  CT2  CE1  X    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        2  3  4  5 1   150  60  0.0 0.0
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={}),
                       Interaction(atoms=(1, 2, 3, 4), parameters=["1", "150", "60", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test priority of defined over generic match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        CT2  CE2  CE1  CT2    5   123.50	401.664	0.0	0.0
        X    CE2  CE1  X      5   20            20      0.0     0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test reverse order for priority of defined over generic match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        X    CE2  CE1  X      5   20            20      0.0     0.0
        CT2  CE2  CE1  CT2    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test generic improper
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        CT2   X     X     CT2    5   123.50	401.664	0.0	0.0
        CT2   X     X     CE2    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 2
        2  3  4  5 2
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={}),
                       Interaction(atoms=(1, 2, 3, 4), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={})]}
        ),
        # multiple matchs and pairs and a meta parameters
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ angletypes ]
        CE1    CE2	CT2	5	123.50	401.664	0.0	0.0
        [ bondtypes ]
        CT2       CE2       1       0.1335  502080.0
        #ifdef old
        CE2       CE1       1       0.1335  502080.0
        #endif
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        [ bonds ]
        1 2 1
        2 3 1
        [ pairs ]
        1  2  1
        [ angles ]
        #ifdef angle
        1  2  3 1
        #endif
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(1, 2), parameters=["1", "0.1335", "502080.0"], meta={'tag': 'old', 'condition': 'ifdef'})],
         "pairs": [Interaction(atoms=(0, 1), parameters=["1"], meta={})],
         "angles": [Interaction(atoms=(0, 1, 2), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={"ifdef":"angle"})]}
        ),
        # test bondtype usage
        (
        """
        #define _FF_OPLS
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        opls_001   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        [ bondtypes ]
        C       C       1       0.1335  502080.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 opls_001   1 test C1 1   0.0 14.0
        2 opls_001   1 test C2 2   0.0 12.0
        [ bonds ]
        1 2  1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={})]}
        ),
        # test virtual_sites n,2,3,4 are skipped
        (
        """
        #define _FF_OPLS
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        opls_001   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        [ bondtypes ]
        C       C       1       0.1335  502080.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 opls_001   1 test C1 1   0.0 14.0
        2 opls_001   1 test C2 2   0.0 12.0
        3 opls_001   1 test C2 2   0.0 12.0
        4 opls_001   1 test C2 2   0.0 12.0
        5 opls_001   1 test C2 2   0.0 12.0
        6 opls_001   1 test C2 2   0.0 12.0
        7 opls_001   1 test C2 2   0.0 12.0
        8 opls_001   1 test C2 2   0.0 12.0
        9 opls_001   1 test C2 2   0.0 12.0
        10 opls_001   1 test C2 2   0.0 12.0
        [ bonds ]
        1  2  1
        1  8  1
        1  9  1
        2  9  1
        8  9  1
        ; currently not parsed accurately due to vermouth bug
        [ virtual_sites2 ]
        4   1  9  1  0.5000
        [ virtual_sites3 ]
        5   4  8  1  1  0.200  0.200
        6   4  9  2  1  0.200  0.200
        [ virtual_sites4 ]
        10   4  8  1  7  1  0.200  0.200  0.300
        [ virtual_sitesn ]
        3   1   4   4   1  2
        7   1   4   4   8  9
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(0, 7), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(0, 8), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(1, 8), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(7, 8), parameters=["1", "0.1335", "502080.0"], meta={})],
         "virtual_sitesn": [Interaction(atoms=(2, 3, 3, 0, 1), parameters=["1"], meta={}),
                            Interaction(atoms=(6, 3, 3, 7, 8), parameters=["1"], meta={})],
         "virtual_sites4": [Interaction(atoms=(9, 3, 7, 0, 6), parameters=["1", "0.200", "0.200", "0.300"], meta={})],
         "virtual_sites2": [Interaction(atoms=(3, 0, 8), parameters=["1", "0.5000"], meta={})],
         "virtual_sites3": [Interaction(atoms=(4, 3, 7, 0), parameters=["1", "0.200", "0.200"], meta={}),
                            Interaction(atoms=(5, 3, 8, 1), parameters=["1", "0.200", "0.200"], meta={})]}
        )
	))
    def test_replace_types(lines, outcome):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        top.gen_bonded_interactions()
        for inter_type in outcome:
            assert top.molecules[0].molecule.interactions[inter_type] == outcome[inter_type]

    @staticmethod
    def test_replace_types_fail():
        lines = """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        C   Q    Q     Q    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        [ dihedrals ]
        1  2  3  4 2
        [ system ]
        some title
        [ molecules ]
        test 1
        """
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        with pytest.raises(OSError):
             top.gen_bonded_interactions()



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import multiprocessing
import numpy as np
import scipy.spatial
from polyply import jit
from .topology import lorentz_berthelot_rule

def _lennard_jones_force(dist, point, ref, params):
    """
    Compute the Lennard-Jones force between two particles
    given their interaction parameters, the distance, and
    distance vector.

    Parameters:
    -----------
    dist: float
        the distance between particles
    vect: np.ndarray(3)
        the distance vector
    params: tuple
        tuple of sigma and epsilon parameters

    Returns:
    --------
    numpy.ndarray(3)
        the force vector
    """
    sig, eps = params
    # distances are computed directly from the KD tree, as we compute the force here
    # we also need the distance vector, which does not come from the KDtree. Computing
    # the distance vector here rather than outside the function profits from the numba
    # acceleration. Computing the distance also here would, however, not be faster
    # as the cKDtree is better optimized still. So we get the distance from outside
    # and compute the vector here. All hail the microoptimization.
    vect = point - ref
    force = 24 * eps / dist * ((2 * (sig/dist)**12.0) - (sig/dist)**6) * vect/dist
    return force

# numba implementation if available
lennard_jones_force = jit(_lennard_jones_force)

POTENTIAL_FUNC = {"LJ": lennard_jones_force}


def _n_particles(molecules):
    """
    Count the number of meta_molecule nodes
    in the topology.
    """
    return sum(map(len, molecules))

class NonBondEngine():
    """
    Stores interactions, positions, and pbc
    conditions in an efficent manner such
    that they can be queried, modified and used to
    compute non-bonded forces in a fast manner.
    The class can be created from the `polyply.src.topology`
    class using the class creation method from_topology.
    """

    def __init__(self,
                 positions,
                 nodes_to_idx,
                 atom_types,
                 interaction_matrix,
                 cut_off,
                 boxsize):
        """
        Parameters:
        -----------
        positions: np.ndarray
        nodes_to_idx: dict[(mol_idx, node_key)]
             correspondance dict between indices in position array
             and a unique node in the system identified by molecule
             index and node_key
        atomtypes:  np.ndarray
             array of atom_types corresponding to the atoms in positions
        interaction_matrix:  dict[frozenset(str, str), tuple(float, float)]
             Dict mapping the atom_types to LJ type interaction parameters,
             that is sigma, epsilon or C6, C12 depending on the potential
             used. Currently only the sigma epsilon form is implemented.
        cut_off: float
             cut-off for which to compute the interaction in nm
        boxsize: np.ndarray
             box dimensions in nm
        """

        self.positions = positions
        self.nodes_to_gndx = nodes_to_idx
        self.atypes = np.asarray(atom_types, dtype=str)
        self.interaction_matrix = interaction_matrix
        self.cut_off = cut_off
        self.boxsize = boxsize

        self.defined_idxs = [list(np.where(self.positions[:, 0].reshape(-1) != np.inf)[0])]
        self.position_trees = [scipy.spatial.ckdtree.cKDTree(positions[self.defined_idxs[-1]],
                                                             boxsize=boxsize)]

        # given a global node index, in which tree is the position saved
        self.gndx_to_tree = {idx: 0 for idx in self.defined_idxs[0]}

    def copy(self):
        """
        Return new instance and deep copy of the objects position attribute.
        """
        new_obj = self.__class__(self.positions.copy(),
                                 self.nodes_to_gndx,
                                 self.atypes,
                                 self.interaction_matrix,
                                 cut_off=self.cut_off,
                                 boxsize=self.boxsize)
        return new_obj

    def concatenate_trees(self):
        """
        Rebuild a single tree from all defined coordinates and remove all
        other trees. This function enables you to condense the positions
        saved over multiple trees into a single tree.
        """
        self.defined_idxs = [list(np.where(self.positions[:, 0].reshape(-1) != np.inf)[0])]
        self.position_trees = [scipy.spatial.ckdtree.cKDTree(self.positions[self.defined_idxs[-1]],
                                                             boxsize=self.boxsize)]
        self.gndx_to_tree = {idx: 0 for idx in self.defined_idxs[0]}

    def add_positions(self, point, mol_idx, node_key, start=True):
        """
        Add `point` with global index `gndx` to the position-matrix
        and position tree.
        """
        gndx = self.nodes_to_gndx[(mol_idx, node_key)]
        self.positions[gndx] = point

        # at around 5000 coordinates it is faster to make a new tree than to add the
        # point to the old tree and profit from faster calculation of distances within
        # a single tree
        if start and self.position_trees[-1].n > 5000:
           self.defined_idxs.append([gndx])
           self.gndx_to_tree[gndx] = len(self.position_trees)
           self.position_trees.append(scipy.spatial.ckdtree.cKDTree(point.reshape(1,3),
                                                                    boxsize=self.boxsize,
                                                                    balanced_tree=False,
                                                                    compact_nodes=False))
        else:
           self.defined_idxs[-1].append(gndx)
           self.gndx_to_tree[gndx] = len(self.position_trees) - 1
           new_tree = scipy.spatial.ckdtree.cKDTree(self.positions[self.defined_idxs[-1]],
                                                    boxsize=self.boxsize,
                                                    balanced_tree=False,
                                                    compact_nodes=False)
           self.position_trees[-1] = new_tree

    def remove_positions(self, mol_idx, node_keys):
        """
        Remove `point` with global index `gndx` from the positions
        matrix and position-tree.

        Paramters
        ---------
        mol_idx: int
            index of the molecule for which to remove_positions
        node_keys: abc.iteratable
            which nodes to remove
        """
        tree_idxs = []
        for node_key in node_keys:
            gndx = self.nodes_to_gndx[(mol_idx, node_key)]
            # guard against when a position is not defined
            if gndx not in self.gndx_to_tree:
               continue
            self.positions[gndx] = np.array([np.inf, np.inf, np.inf])
            tree_idx = self.gndx_to_tree[gndx]
            self.defined_idxs[tree_idx].remove(gndx)
            del self.gndx_to_tree[gndx]
            tree_idxs.append(tree_idx)

        for tree_idx in tree_idxs:
            new_tree = scipy.spatial.ckdtree.cKDTree(self.positions[self.defined_idxs[tree_idx]],
                                                     boxsize=self.boxsize,
                                                     balanced_tree=False,
                                                     compact_nodes=False)
            self.position_trees[tree_idx] = new_tree

    def update_positions_in_molecules(self, molecules):
        """
        Add the positions stored in the object back
        to the molecules.
        """
        for mol_idx, molecule in enumerate(molecules):
            for node in molecule.nodes:
                gndx = self.nodes_to_gndx[(mol_idx, node)]
                molecule.nodes[node]["position"] = self.positions[gndx]

    def get_point(self, mol_idx, node):
        """
        Get the position of a point from the molecule index
        and node_key.
        """
        gndx = self.nodes_to_gndx[(mol_idx, node)]
        return self.positions[gndx]

    def get_interaction(self, mol_idx_a, mol_idx_b, node_a, node_b):
        """
        Get the interaction parameters between two nodes identified
        by their molecule indices and node-keys.
        """
        gndx_a = self.nodes_to_gndx[(mol_idx_a, node_a)]
        gndx_b = self.nodes_to_gndx[(mol_idx_b, node_b)]
        atype_a = self.atypes[gndx_a]
        atype_b = self.atypes[gndx_b]
        return self.interaction_matrix[frozenset([atype_a, atype_b])]

    def compute_force_point(self, point, mol_idx, node, exclude=[], potential="LJ"):
        """
        Compute the force on `node` of molecule `mol_idx` with coordinates
        `point` given the potential definition in `potential`.

        Parameters
        ----------
        point: np.ndarray(3)
        mol_idx: int
        node: abc.hashable
            node key
        exclude: list[collections.abc.hashable]
            list of nodes to exclude from computation
        potential: abc.hashable
            definition of potential to use

        Returns
        -------
        np.ndarray(3)
            the force vector
        """
        exclusions = [self.nodes_to_gndx[(mol_idx, node)] for node in exclude]

        ref_tree = scipy.spatial.ckdtree.cKDTree(point.reshape(1, 3),
                                                 boxsize=self.boxsize)
        force = 0
        for pos_tree, defined_idxs in zip(self.position_trees, self.defined_idxs):

            dist_mat = ref_tree.sparse_distance_matrix(pos_tree, self.cut_off)

            if any(np.array(list(dist_mat.values())) < 0.1):
               return np.inf

            gndx = self.nodes_to_gndx[(mol_idx, node)]
            current_atype = self.atypes[gndx]

            for pair, dist in dist_mat.items():
                gndx_pair = defined_idxs[pair[1]]

                if gndx_pair not in exclusions:
                    other_atype = self.atypes[gndx_pair]
                    params = self.interaction_matrix[frozenset([current_atype, other_atype])]
                    force += POTENTIAL_FUNC[potential](dist, point, self.positions[gndx_pair], params)
        return force

    @classmethod
    def from_topology(cls, molecules, topology, box):
        """
        Create a class instance from a topology object,
        a list of molecules and a box.

        Parameters:
        -----------
        molecules: list
        topology: :class:`polyply.src.topology`
        box: np.nadarray
        """

        n_atoms = _n_particles(molecules)

        # array of all positions
        positions = np.ones((n_atoms, 3)) * np.inf
        # convert molecule index and node to index
        # in position matrix
        nodes_to_gndx = {}

        atom_types = []
        idx = 0
        mol_count = 0
        for molecule in molecules:
            for node in molecule.nodes:
                if "position" in molecule.nodes[node]:
                    positions[idx, :] = molecule.nodes[node]["position"]

                resname = molecule.nodes[node]["resname"]
                atom_types.append(resname)
                nodes_to_gndx[(mol_count, node)] = idx
                idx += 1

            mol_count += 1

        inter_matrix = {}
        for res_a, res_b in itertools.combinations(set(atom_types), r=2):
            # as by model definition all epsilon values are set to 1
            params = lorentz_berthelot_rule(topology.volumes[res_a],
                                            topology.volumes[res_b], 1.0, 1.0)
            inter_matrix[frozenset([res_a, res_b])] = params

        for resname in set(atom_types):
            vdw_radius = topology.volumes[resname]
            # as by model definition all epsilon values are set to 1
            inter_matrix[frozenset([resname, resname])] = (vdw_radius, 1.0)

        # dynamically set the cut-off as twice the largest vdw-radius
        cut_off = max(list(inter_matrix.values()))[0] * 2.
        nonbond_matrix = cls(positions, nodes_to_gndx,
                             atom_types, inter_matrix, cut_off=cut_off, boxsize=box)
        return nonbond_matrix



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

# TODO:
# add test for when version tagging is important

import textwrap
import pytest
import numpy as np
import vermouth.forcefield
import vermouth.molecule
import polyply.src.polyply_parser

class TestPolyply:
    @staticmethod
    def test_dangling_bond():
        lines = """
        [ moleculetype ]
        ; name nexcl.
        PEO         1
        ;
        [ atoms ]
        1  SN1a    1   PEO   COC  1   0.000  45
        [ bonds ]
        ; back bone bonds
        1  2   1   0.37 7000
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        polyply.src.polyply_parser.read_polyply(lines, ff)

        link_bond = vermouth.molecule.Interaction(
                atoms=["COC", "+COC"], parameters=['1', '0.37', '7000'], meta={"version":1},)
        assert ff.blocks['PEO'].interactions['bonds'] == []
        assert ff.links[0].interactions['bonds'][0] == link_bond

    @staticmethod
    def test_multiple_interactions():
        lines = """
        [ moleculetype ]
        ; name nexcl.
        PEO         1
        ;
        [ atoms ]
        1  SN1a    1   PEO   COC  1   0.000  45
        [ dihedrals ]
        1  2  3  4     1    180.00    1.96   1
        1  2  3  4     1     0        0.18   2
        1  2  3  4     1     0        0.33   3
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        polyply.src.polyply_parser.read_polyply(lines, ff)

        dih  = [vermouth.molecule.Interaction(
                atoms=["COC", "+COC", "++COC", "+++COC"], parameters=['1','180.00','1.96','1'], meta={'version':3}),
               vermouth.molecule.Interaction(
                atoms=["COC", "+COC", "++COC", "+++COC"], parameters=['1','0','0.18','2'], meta={'version':2}),
               vermouth.molecule.Interaction(
                atoms=["COC", "+COC", "++COC", "+++COC"], parameters=['1','0','0.33','3'], meta={'version':1})]

        assert len(ff.links) == 1
        assert ff.links[0].interactions['dihedrals'] == dih

    @staticmethod
    def test_exclusions():
        lines = """
        [ moleculetype ]
        ; name nexcl.
        PEO         1
        ;
        [ atoms ]
        1  SN1a    1   PEO   COC  1   0.000  45
        [ exclusions ]
        1  3
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        polyply.src.polyply_parser.read_polyply(lines, ff)

        excl = [vermouth.molecule.Interaction(
                atoms=["COC", "++COC"], parameters=[], meta={"version":1},)]

        assert ff.links[0].interactions['exclusions'] == excl


    @staticmethod
    @pytest.mark.parametrize('lines, total, blocks, links', (
        ("""
        [ moleculetype ]
        ; name nexcl.
        PEO         1
        ;
        [ atoms ]
        1  SN1a    1   PEO   EO1  1   0.000  45
        2  SN1a    1   PEO   EO2  1   0.000  45
        [ bonds ]
        ; back bone bonds
        1  2   1   0.37 7000
        2  3   1   0.37 7000
        """,
        1,
        {"bonds":[ vermouth.molecule.Interaction(
                atoms=[0, 1], parameters=['1', '0.37', '7000'], meta={},)]},
        {"bonds":[vermouth.molecule.Interaction(
                atoms=["EO2", "+EO1"], parameters=['1', '0.37', '7000'], meta={"version":1},)]}
        ),
        ("""
         [ moleculetype ]
         PS 1
         [ atoms ]
            1    STY            1  PS       R1       1     0.00000E+00   45
            2    STY            1  PS       R2       2     0.00000E+00   45
            3    STY            1  PS       R3       3     0.00000E+00   45
            4    SCY            1  PS       B        4     0.00000E+00   45
         [ bonds ]
            1     4   1     0.27 8000
            4     5   1     0.27 8000

         [ angles ]
           4     1     2    1   136  100
           4     1     3    1   136  100
           ; links
           4     5     6    1   136  100
           4     5     7    1   136  100
           1     4     5    1   120   25
           4     5     8    1    52  550
        """,
        5,
        {"bonds": [vermouth.molecule.Interaction(
                atoms=[0, 3], parameters=['1', '0.27', '8000'], meta={},)],
         "angles":[vermouth.molecule.Interaction(
                atoms=[3, 0, 1], parameters=['1', '136', '100'], meta={},),
                   vermouth.molecule.Interaction(
                atoms=[3, 0, 2], parameters=['1', '136', '100'], meta={},)]},
        {"bonds": [vermouth.molecule.Interaction(
                atoms=["B", "+R1"], parameters=['1', '0.27', '8000'], meta={'version':1},)],
         "angles":[
                vermouth.molecule.Interaction(
                atoms=["B", "+R1", "+R2"], parameters=['1', '136', '100'], meta={'version':1},),
                vermouth.molecule.Interaction(
                atoms=["B", "+R1", "+R2"], parameters=['1', '136', '100'], meta={'version':1},),
                vermouth.molecule.Interaction(
                atoms=["R1", "B", "+R1"], parameters=['1', '120', '25'], meta={'version':1},),
                vermouth.molecule.Interaction(
                atoms=["B", "+R1", "+B"], parameters=['1', '52', '550'], meta={'version':1},)]},
        )))

    def test_split_link_block(lines, total, blocks, links):
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        polyply.src.polyply_parser.read_polyply(lines, ff)
        # test the correct total number of links is produced
        assert len(ff.links) == total
        # check if each link has the length one
        for link in ff.links:
            assert len(link.interactions) == 1
            key = list(link.interactions.keys())[0]
            values = link.interactions[key]
            print(values)
            assert len(values) == 1
            assert len(link.edges) != 0



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Collection of useful functions for performing operations based
on graphs.
"""
from collections import defaultdict
import networkx as nx

def neighborhood(graph, source, max_length, min_length=1):
    """
    Returns all neighbours of `source` that are less or equal
    to `cutoff` nodes away and more or equal to `start` away
    within a graph excluding the node itself.

    Parameters
    ----------
    graph: :class:`networkx.Graph`
        A networkx graph definintion
    source:
        A node key matching one in the graph
    max_length: :type:`int`
        The maxdistance between the node and its
        neighbours.
    min_length: :type:`int`
        The minimum length of a path. Default
        is zero

    Returns
    --------
    list
       list of all nodes distance away from reference
    """
    paths = nx.single_source_shortest_path(G=graph, source=source, cutoff=max_length)
    neighbours = [node for node, path in paths.items() if min_length <= len(path)]
    return neighbours

def find_nodes_with_attributes(graph, **attrs):
    """
    Yields all nodes in graph, which have all
    attributes in attrs set to value.

    Parameters:
    -----------
    graph: nx.Graph
    **attrs: collections.abc.Mapping
        The attributes and their desired values.

    Yields:
    --------
    collections.abc.Hashable
    """
    for node in graph.nodes:
        for attr, value in attrs.items():
            if attr not in graph.nodes[node] or graph.nodes[node][attr] != value:
                break
        else:
            yield node

def is_branched(graph):
    """
    Check if any node has a degree larger than 2

    Parameters:
    -----------
    graph: :class:`networkx.Graph`
        A networkx graph definintion

    Returns:
    --------
    bool
       is branched
    """
    for _, deg in graph.degree:
        if deg > 2:
            return True
    return False

def find_connecting_edges(res_graph, molecule, nodes):
    """
    Given a list of `nodes` referring to specific nodes in a
    molecule residue graph (`res_graph`) find all edges in the
    molecule graph, which connect the two residues. This function
    avoids looping over all edges in the molecule graph, so it
    scales well.

    Parameters
    ----------
    res_graph: :class:`nx.Graph`
        residue graph; must have the attribute "graph" and
        "resid", where graph is the fragment the node describes
        in the mol_graph
    molecule: :class:`vermouth.molecule.Molecule`
        vermouth molecule underlying the residue graph
    nodes: (node_key, node_key)
        tuple of node-keys specifing a residue in residue graph

    Returns
    ----------
    list
       list of edges found
    """
    # First find all nodes in the residue graph fragments,
    # whose degree is unequal to their degree in complete
    # the molecule. This is an efficent way to filter which
    # atoms are involved in a potential link, because it reduces
    # the search space to only those atoms that have a dangling
    # edge in the residue as compared to the complete molecule.
    # Then sotre these nodes together with the residue they are
    # found in.
    allowed_nodes = defaultdict(list)
    for res_node in nodes:
        for node in res_graph.nodes[res_node]["graph"].nodes:
            deg = res_graph.nodes[res_node]["graph"].degree(node)
            if deg != molecule.degree(node):
                allowed_nodes[res_node].append(node)

    # given all nodes, which potentially have a connecting edge
    # we only need to check generate all possible edges and check
    # if they exist
    edges = []
    for high_res_node_a in allowed_nodes[nodes[0]]:
        for high_res_node_b in allowed_nodes[nodes[1]]:
            if molecule.has_edge(high_res_node_a, high_res_node_b):
                edges.append((high_res_node_a, high_res_node_b))

    return edges



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# this processor is not a subclass to the regular processor
# class because it cannot run on a single molecule but needs
# the system information
"""
Processor for building systems with more than one molecule
"""
import inspect
import numpy as np
from tqdm import tqdm
from .random_walk import RandomWalk
from .linalg_functions import norm_sphere
from .nonbond_engine import NonBondEngine

def _compute_box_size(topology, density):
    """
    Lookup the masses in the `topology`
    and compute a cubic box that matches
    the `density` given the number of molecules
    defined in the topology. Units are nm

    Parameters:
    -----------
    topology: :class:`polyply.src.topology`
    density: float
       target density

    Returns:
    --------
    float
        the edge length of cubix box
    """
    total_mass = 0
    for meta_molecule in topology.molecules:
        molecule = meta_molecule.molecule
        for node in molecule.nodes:
            if 'mass' in molecule.nodes[node]:
                total_mass += molecule.nodes[node]['mass']
            else:
                try:
                    atype = molecule.nodes[node]["atype"]
                    total_mass += topology.atom_types[atype]['mass']
                except KeyError as error:
                    msg = ("Trying to compute system density, but cannot "
                           "find mass of atom {} with type {} in topology.")
                    atom = molecule.nodes[node]["atomname"]
                    raise KeyError(msg.format(atom, atype)) from error

    # amu -> kg and cm3 -> nm3
    # conversion = 1.6605410*10**-27 * 10**27
    box = (total_mass*1.6605410/density)**(1/3.)
    return box

def _filter_by_molname(molecules, ignore):
    """
    Selects all molecules from `molecules` which
    do not have a name mentioned in `ignore`.
    """
    for molecule in molecules:
        if molecule.mol_name not in ignore:
            yield molecule

class BuildSystem():
    """
    Compose a system of molecules according
    to the definitions in the topology file.

    This class when run for a system or a molecule
    calls the random-walk processor to generate the
    coordinates for a single moleccule. In contrast
    to the random-walk processor this class handles
    all system related accouting matters such as
    starting points, box-dimensions and existing
    molecules and so forth.

    Parameters:
    -----------
    topology: :class:`polyply.src.topology`
    density: float
        the system density
    start_dict: dict
        a dictionary associating ...
    grid_spacing: flaot
        the distance between grid points
    grid: np.ndarray
        a grid defining grid points this argument
        overwrides generation of the grid using
        the grid-spacing
    maxiter: int
        maximum number of tries to genrate a single
        molecule
    box: np.ndarray[3,1]
        box size, this overwrites generation of the box
        by the density argument
    ignore: list[str]
        list of molecule names to ignore when building
    **kwargs:
        all passed down to random-walk
    """

    def __init__(self,
                 topology,
                 density,
                 start_dict,
                 grid_spacing=0.2,
                 maxiter=800,
                 box=None,
                 ignore=[],
                 grid=None,
                 **kwargs):

        self.topology = topology
        self.density = density
        self.grid_spacing = grid_spacing
        self.maxiter = maxiter
        self.ignore = ignore
        self.box_grid = grid
        self.box = box
        self.start_dict = start_dict

        # first we check if **kwargs are actually in random-walk
        valid_kwargs = inspect.getfullargspec(RandomWalk).args
        for kwarg in kwargs:
            if kwarg not in valid_kwargs:
               msg = ("Keyword argument {} is not valid for the "
                      "RandomWalk processor class. ")
               raise TypeError(msg.format(kwarg))
        self.rwargs = kwargs

        # set the box if a box is given
        # we use type comparison because box is an array
        if not isinstance(self.box, type(None)):
            self.box = box
        # if box is not given but density compute it from density
        else:
            box_dim = round(_compute_box_size(topology, self.density), 5)
            self.box = np.array([box_dim, box_dim, box_dim])

        # intialize the grid if there is none given
        if isinstance(self.box_grid, type(None)):
            self.box_grid = np.mgrid[0:self.box[0]:self.grid_spacing,
                                     0:self.box[1]:self.grid_spacing,
                                     0:self.box[2]:self.grid_spacing].reshape(3, -1).T

        # this should be done elsewhere
        topology.box = (self.box[0], self.box[1], self.box[2])

    def _handle_random_walk(self, molecule, mol_idx, vector_sphere):
        step_count = 0
        while True:
            start_idx = np.random.randint(len(self.box_grid))
            start = self.box_grid[start_idx]

            processor = RandomWalk(mol_idx,
                                   self.nonbond_matrix,
                                   start=start,
                                   maxdim=self.box,
                                   vector_sphere=vector_sphere,
                                   start_node=self.start_dict[mol_idx],
                                   **self.rwargs)

            processor.run_molecule(molecule)

            if processor.success:
                return True, processor.nonbond_matrix
            elif step_count == self.maxiter:
                processor.nonbond_matrix.remove_positions(mol_idx, molecule.nodes)
                return False, processor.nonbond_matrix
            else:
                step_count += 1
                self.nonbond_matrix.remove_positions(mol_idx, molecule.nodes)

    def _compose_system(self, molecules):
        """
        Place the molecules of the system into a box
        and optimize positions to meet density.

        Parameters
        ----------
        topology:  :class:`vermouth.system`
        density: foat
           density of the system in kg/cm3

        Returns
        --------
        system
        """
        mol_idx = 0
        pbar = tqdm(total=len(molecules))
        mol_tot = len(molecules)
        vector_sphere = norm_sphere(5000)
        while mol_idx < mol_tot:

            molecule = molecules[mol_idx]

            if all(["position" in molecule.nodes[node] for node in molecule.nodes]):
                mol_idx += 1
                pbar.update(1)
                continue

            success, new_nonbond_matrix = self._handle_random_walk(molecule,
                                                                   mol_idx,
                                                                   vector_sphere)
            if success:
                self.nonbond_matrix = new_nonbond_matrix
                self.nonbond_matrix.concatenate_trees()
                mol_idx += 1
                pbar.update(1)

        pbar.close()
        self.nonbond_matrix.update_positions_in_molecules(molecules)

    def run_system(self, molecules):
        """
        Compose a system according to a the system
        specifications and a density value.
        """
        # filter all molecules that should be ignored during the building process
        self.molecules = list(_filter_by_molname(self.topology.molecules, self.ignore))

        # generate the nonbonded matrix wrapping all information about molecular
        # interactions
        self.nonbond_matrix = NonBondEngine.from_topology(self.molecules, self.topology, self.box)
        self._compose_system(self.molecules)
        return molecules



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Runs more elaborate integration tests
"""

from collections import defaultdict
from pathlib import Path
import shlex
import subprocess
import sys
import numpy as np

import pytest
import vermouth
from vermouth.forcefield import ForceField

from polyply import TEST_DATA
from vermouth.tests.helper_functions import find_in_path

INTEGRATION_DATA = Path(TEST_DATA + '/library_tests')

PATTERN = '{path}/{library}/{polymer}/polyply'

POLYPLY = find_in_path(names=("polyply", ))

def assert_equal_blocks(block1, block2):
    """
    Asserts that two blocks are equal to gain the pytest rich comparisons,
    which is lost when doing `assert block1 == block2`
    """
    assert block1.name == block2.name
    assert block1.nrexcl == block2.nrexcl
    assert block1.force_field == block2.force_field  # Set to be equal

    for node in block1.nodes:
        # for the simulation only these two attributes matter
        # as we have 3rd party reference files we don't do more
        # checks
        for attr in ["atype", "charge"]:
            assert block1.nodes[node][attr] == block2.nodes[node][attr]

    edges1 = {frozenset(e[:2]): e[2] for e in block1.edges(data=True)}
    edges2 = {frozenset(e[:2]): e[2] for e in block2.edges(data=True)}

    for e, attrs in edges2.items():
        for k, v in attrs.items():
            if isinstance(v, float):
                attrs[k] = pytest.approx(v, abs=1e-3) # PDB precision is 1e-3

    assert edges1 == edges2

    for inter_type in ["bonds", "angles", "constraints", "exclusions", "pairs", "dihedrals", "impropers"]:
        ref_interactions = block1.interactions.get(inter_type, [])
        new_interactions = block2.interactions.get(inter_type, [])
        print(inter_type)
        assert len(ref_interactions) == len(new_interactions)

        ref_terms = defaultdict(list)
        for inter in ref_interactions:
            atoms = inter.atoms
            ref_terms[frozenset(atoms)].append(inter)

        new_terms = defaultdict(list)
        for inter_new in new_interactions:
            atoms = inter_new.atoms
            new_terms[frozenset(atoms)].append(inter_new)

        for atoms, ref_interactions in ref_terms.items():
            new_interactions = new_terms[atoms]
            for ref_inter in ref_interactions:
                for new_inter in new_interactions:
                    if _interaction_equal(ref_inter, new_inter, inter_type):
                        break
                else:
                    assert False

def compare_itp(filename1, filename2):
    """
    Asserts that two itps are functionally identical
    """
    dummy_ff = ForceField(name='dummy')
    with open(filename1) as fn1:
        vermouth.gmx.read_itp(fn1, dummy_ff)
    dummy_ff2 = ForceField(name='dummy')
    with open(filename2) as fn2:
        vermouth.gmx.read_itp(fn2, dummy_ff2)
    for block in dummy_ff2.blocks.values():
        block._force_field = dummy_ff
    assert set(dummy_ff.blocks.keys()) == set(dummy_ff2.blocks.keys())
    for name in dummy_ff.blocks:
        block1 = dummy_ff.blocks[name]
        block2 = dummy_ff2.blocks[name]
        assert_equal_blocks(block1, block2)


def compare_pdb(filename1, filename2):
    """
    Asserts that two pdbs are functionally identical
    """
    pdb1 = vermouth.pdb.read_pdb(filename1)
    pdb2 = vermouth.pdb.read_pdb(filename2)
    assert len(pdb1) == len(pdb2)
    for mol1, mol2 in zip(pdb1, pdb2):
        for mol in (mol1, mol2):
            for n_idx in mol:
                node = mol.nodes[n_idx]
                if 'position' in node and node['atomname'] in ('SCN', 'SCP'):
                    # Charge dummies get placed randomly, which complicated
                    # comparisons to no end.
                    # These will be caught by the distances in the edges instead.
                    del node['position']

        assert_equal_blocks(mol1, mol2)


COMPARERS = {'.itp': compare_itp,
             '.pdb': compare_pdb}


def _interaction_equal(interaction1, interaction2, inter_type):
    """
    Returns True if interaction1 == interaction2, ignoring rounding errors in
    interaction parameters.
    """
    p1 = list(map(str, interaction1.parameters))
    p2 = list(map(str, interaction2.parameters))
    a1 = list(interaction1.atoms)
    a2 = list(interaction2.atoms)


    if p1 != p2:
        return False

    if interaction1.meta != interaction2.meta:
        return False

    if inter_type in ["constraints", "bonds", "exclusions", "pairs"]:
        a1.sort()
        a2.sort()
        return a1 == a2

    elif inter_type in ["impropers", "dihedrals"]:
         if frozenset([a1[0], a1[1], a1[2]]) == frozenset([a2[0], a2[1], a2[2]]) and a1[3] == a2[3]:
            return True
         elif frozenset([a1[0], a1[1], a1[2]]) == frozenset([a2[1], a2[2], a2[3]]) and a1[3] == a2[0]:
            return True
         elif frozenset([a1[1], a1[2], a1[3]]) == frozenset([a2[1], a2[2], a2[3]]) and a1[0] == a2[0]:
            return True

    elif inter_type in ["angles"]:
        return a1[1] == a2[1] and frozenset([a1[0], a1[2]]) == frozenset([a2[0], a2[2]])

    return False

@pytest.mark.parametrize("library, polymer", [
     ['2016H66', 'PP'],
     ['2016H66', 'C12E4'],
     ['2016H66', 'PE'],
     ['2016H66', 'PVA'],
     ['2016H66', 'PMA'],
     ['2016H66', 'PS'],
     ['gromos53A6', 'P3HT'],
     ['martini3', 'PEO'],
     ['martini3', 'PS'],
     ['martini3', 'PE'],
     ['martini3', 'DEX'],
     ['martini3', 'P3HT'],
     ['martini2', 'PEO'],
     ['martini2', 'PS'],
     ['martini2', 'PEL'],
     ['martini2', 'PEO_PE'],
  # -> proteins?
])
def test_integration_protein(tmp_path, monkeypatch, library, polymer):
    """
    Runs tests on the library by executing the contents of the file `command` in
    the folder libname/polymer, and tests whether the contents of the produced
    files are the same as the reference files. The comparison of the files is
    governed by `COMPARERS`.

    Parameters
    ----------
    tmp_path
    library: str
    polymer: str
    """
    monkeypatch.chdir(tmp_path)
    data_path = Path(PATTERN.format(path=INTEGRATION_DATA, library=library, polymer=polymer))

    with open(str(data_path / 'command')) as cmd_file:
        command = cmd_file.read().strip()

    assert command  # Defensive
    command = shlex.split(command)
    result = [sys.executable]
    for token in command:
        if token.startswith('polyply'):
            result.append(str(POLYPLY))
        elif token.startswith('.'):
            result.append(str(data_path / token))
        else:
            result.append(token)

    command = result

    # read the citations that are expected
    # citations = []
    # with open(str(data_path/'citation')) as cite_file:
    #     for line in cite_file:
    #         citations.append(line.strip())

    proc = subprocess.run(command, cwd='.', timeout=60, check=False,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          universal_newlines=True)
    exit_code = proc.returncode
    if exit_code:
        print(proc.stdout)
        print(proc.stderr)
        assert not exit_code

    # check if strdout has citations in string
    #for citation in citations:
    #    assert citation in proc.stderr

    files = list(tmp_path.iterdir())

    assert files
    assert list(tmp_path.glob('*.itp')), files

    for new_file in tmp_path.iterdir():
        filename = new_file.name
        reference_file = data_path/filename
        assert reference_file.is_file()
        ext = new_file.suffix.lower()
        if ext in COMPARERS:
            COMPARERS[ext](str(reference_file), str(new_file))



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import numpy as np
import scipy
import scipy.optimize
from .linalg_functions import (angle, dih)
from .virtual_site_builder import construct_vs
"""
Processor and functions for optimizing the geomtry
of vermoth molecules.
"""
# we use weights in the minization procedure
# to account for the fact that differences in
# bond length are typically small owing to the
# fact that the length unit is nm. Scaling them
# by 10,000 also increases their weight over
# angles, and dihedrals, which is wanted as
# the bonds have to be very close to their
# equilbrium value
WEIGHTS = {"bonds": 10000.0,
           "angles": 1.0,
           "constraints": 10000.0,
           "dihedrals": 1.0,}

def compute_bond(params, coords):
    """
    Compute the distance between two points in `coords`
    and then take the MSD with repsect to a reference
    value provided in `params` multiplyed by 1000. The
    factor 1000 enforces that the bounds get a higher
    weight in optimization than angles and dihedrals,
    as we deal in units of nm.

    Parameters
    -----------
    params:   list
    coodrs:    numpy.ndarray

    Returns
    -------
    float
    """
    dist = np.linalg.norm(coords[0] - coords[1])
    return WEIGHTS["bonds"] *(dist - float(params[1]))**2.0

def compute_angle(params, coords):
    """
    Compute the angle between three points in `coords`
    and then take the MSD with respect to a reference
    value provided in `params`.

    Parameters
    -----------
    params:   list
    coords:    numpy.ndarray

    Returns
    -------
    float
    """
    angle_value = angle(coords[0], coords[1], coords[2])
    return WEIGHTS["angles"] * (angle_value - float(params[1]))**2.0

def compute_improper_dih(params, coords):
    """
    Compute the improper dihedral angle of GROMACS type 2
    between four points in `coords` and then take the MSD
    with repsect to a reference value provided in `params`.

    Parameters
    -----------
    params:   list
    coords:    numpy.ndarray

    Returns
    -------
    float
    """
    # we filter all non-imporpers out here because
    # the itp file parser doesn't distinguish them
    if params[0] == "2":
        dih_angle = dih(coords[0], coords[1], coords[2], coords[3])
        penalty = WEIGHTS["dihedrals"] * (dih_angle - float(params[1]))**2.0
    else:
        penalty = 0
    return penalty

def renew_vs(positions, block, atom_to_idx):
    """
    Given `positions` update the virtual-sites
    that are specified in block using the current
    positions of the other atoms.

    Parameters
    ----------
    positions: dict
    block: :class:vermouth.molecule.Block

    Returns
    -------
    positions
      the positions with new vs coordinates
    """
    vs_types = ["virtual_sitesn", "virtual_sites2", "virtual_sites3", "virtual_sites4"]
    for vs_type in vs_types:
        interactions = block.interactions.get(vs_type, [])
        for virtual_site in interactions:
            vs_tb = atom_to_idx[virtual_site.atoms[0]]
            pos_dict = {}
            for atom in virtual_site.atoms:
                pos_dict[atom] = positions[atom_to_idx[atom]]
            new_vs = construct_vs(vs_type, virtual_site, pos_dict)
            positions[vs_tb] = new_vs
    return positions


INTER_METHODS = {"bonds": compute_bond,
                 "constraints": compute_bond,
                 "angles": compute_angle,
                 "dihedrals": compute_improper_dih}
# Note that we only optimize improper dihedral angles
# because they keep molecules flat, whereas proper
# dihedral angles typically are fixed well in the
# energy minimization. We could optimize proper
# dihedral angles but then we'd also have to support
# all different function types and their multipicity.

def optimize_geometry(block, coords, inter_types=[], tolerance={"angles": 5,
                                                                "dihedrals": 5,
                                                                "bonds": 0.05,
                                                                "constraints": 0.05}):
    """
    Take the definitions of a `block` and associated
    `coords` and optimize the geometry based on the
    bonds, angles and dihedrals provided in the
    block definition.

    Parameters
    ----------
    block:  :class:vermouth.molecule.Block
    coords: dict
        dictionary of coordinates in form atom_name:np.ndarray
    tolerance: dict
        dictionary of allowed tolerance in nm or degree per
        interaction-type

    Returns
    -------
    bool
      status of the optimization i.e. failure or success
    dict
      dictionary of optimized coordinates
    """
    n_atoms = len(coords)
    atom_to_idx = OrderedDict(zip(list(coords.keys()), range(0, n_atoms)))
    positions = np.array(list(coords.values()))

    def target_function(positions):
        energy = 0
        positions = positions.reshape((-1, 3))
        positions = renew_vs(positions, block, atom_to_idx)
        for inter_type in inter_types:
            interactions = block.interactions.get(inter_type, [])
            for interaction in interactions:
                atoms = interaction.atoms
                params = interaction.parameters
                atom_coords = [positions[atom_to_idx[name]]
                               for name in atoms]
                new = INTER_METHODS[inter_type](params, atom_coords)
                energy += new
        return energy

    opt_results = scipy.optimize.minimize(target_function, positions, method='L-BFGS-B',
                                          options={'ftol': 0.001, 'maxiter': 100})

    # evaluate if optimization was successful
    positions = opt_results['x'].reshape((-1, 3))
    positions = renew_vs(positions, block, atom_to_idx)

    for node_key, idx in atom_to_idx.items():
        coords[node_key] = positions[idx]

    for inter_type in inter_types:
        interactions = block.interactions.get(inter_type, [])
        for interaction in interactions:
            atoms = interaction.atoms
            params = interaction.parameters
            atom_coords = [positions[atom_to_idx[name]]
                           for name in atoms]
            penalty = INTER_METHODS[inter_type](params, atom_coords)
            if penalty > WEIGHTS[inter_type]*tolerance[inter_type]**2.:
                return False, coords

    return True, coords



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""
import textwrap
import pytest
import networkx as nx
import vermouth.forcefield
import vermouth.ffinput
from vermouth.molecule import Block, Molecule, Interaction, Link
import polyply
from polyply.src.meta_molecule import (MetaMolecule, Monomer)
from polyply.src.apply_links import MatchError, ApplyLinks


@pytest.mark.parametrize('resids, order, expected',(
                        ([1, 1, 1], [1, 1, 1], True),
                        ([1, 2, 3], [1, 1, 1], False),
                        ([1, 2, 3], [0, 1, 2], True),
                        ([1, 2, 2], [0, 1, 1], True),
                        ([1, 2, 3], [0, '>', '>>'], True),
                        ([1, 2, 3], [0, '>', '>'], False),
                        ([1, 2, 3], ['<', '<<', 0], False),
                        ([1, 2, 3], ['<<', '<', 0], True),
                        ([1, 2, 3], ['<', '<', 0], False)))
def test_check_relative_orders(resids, order, expected):
    status = polyply.src.apply_links._check_relative_order(resids, order)
    assert status == expected

@pytest.mark.parametrize('attributes, ignore, expected',(
                        ({'resid': 1}, [], [0, 1]),
                        ({'resid': 1, 'type': 'A'}, [], [0]),
                        ({'type': 'A'}, [], [0, 2]),
                        ({}, [], [0, 1, 2, 3]),
                        ({'name': 'C'}, [], [2, 3]),
                        ({'order': 1, 'type': 'D'}, ['type'], [2, 3]),
                        ({'resid': 2, 'name': 'B'}, ['resid'], [1]),
                        ({'order': 0, 'name': 'A'}, ['name'], [0, 1])))
def test_find_atoms(attributes, ignore, expected):
    molecule = nx.Graph()
    molecule.add_nodes_from([(0 , {'resid': 1, 'order': 0, 'name': 'A', 'type': 'A'}),
                            (1 , {'resid': 1, 'order': 0, 'name': 'B', 'type': 'B'}),
                            (2 , {'resid': 2, 'order': 1, 'name': 'C', 'type': 'A'}),
                            (3 , {'resid': 3, 'order': 1, 'name': 'C', 'type': 'C'})])
    atoms = polyply.src.apply_links.find_atoms(molecule, ignore=ignore, **attributes)
    assert set(atoms) == set(expected)

@pytest.fixture
def example_meta_molecule():
    """
    Example molecule with three residues each with
    different attributes for testing the assignment
    of links.

    Names:
    -------
    BB - BB1 - BB - BB1 - BB2 - BB - BB1
          |          |                |
         SC1        SC1              SC1
          |          |                |
         SC2        SC2              SC2

    Nodes:
    ------
     0  - 1    4  - 5 - 6   9 - 10
          |         |           |
          2         7           11
          |         |           |
          3         8           12
    """
    force_field = vermouth.forcefield.ForceField("test")
    block_A = Block(force_field=force_field)
    block_A.add_nodes_from([(0 , {'resid': 1,  'name': 'BB',  'atype': 'A', 'charge': 0.0, 'other': 'A'}),
                            (1 , {'resid': 1,  'name': 'BB1', 'atype': 'B', 'charge': 0.0, 'other': 'A'}),
                            (2 , {'resid': 1,  'name': 'SC1', 'atype': 'C', 'charge': 0.0, 'other': 'A'}),
                            (3 , {'resid': 1,  'name': 'SC2', 'atype': 'D', 'charge': 1.0, 'other': 'A'})])
    block_A.add_edges_from([(0, 1), (1, 2), (2, 3)])

    block_B = Block(force_field=force_field)
    block_B.add_nodes_from([(0 , {'resid': 1,  'name': 'BB', 'atype': 'A', 'charge': 0.0}),
                            (1 , {'resid': 1,  'name': 'BB1', 'atype': 'B', 'charge': 0.0}),
                            (2 , {'resid': 1,  'name': 'BB2', 'atype': 'A', 'charge': 0.0}),
                            (3 , {'resid': 1,  'name': 'SC1', 'atype': 'A', 'charge': -0.5}),
                            (4 , {'resid': 1,  'name': 'SC2', 'atype': 'C', 'charge': 0.5})])
    block_B.add_edges_from([(0, 1), (1, 2), (2, 3), (1, 4)])

    molecule = block_A.to_molecule()
    molecule.merge_molecule(block_B)
    molecule.merge_molecule(block_A)
    molecule.add_edges_from([(1, 4), (8, 9)])

    graph = MetaMolecule._block_graph_to_res_graph(molecule)
    meta_mol = MetaMolecule(graph, force_field=force_field, mol_name="test")
    # before do links is called there are no edges between the residues
    # at lower level
    molecule.remove_edges_from([(1, 4), (8, 9)])
    meta_mol.molecule = molecule
    return meta_mol

@pytest.mark.parametrize('link_nodes, link_to_resid, expected',(
                        # first and second residue name-based
                        ([(0, {'name': 'BB1'}),
                          (1, {'name': 'BB'})],
                          {0: 0, 1: 1},
                          {0: 1, 1: 4}
                        ),
                        # second third residue name-based
                        ([(0, {'name': 'BB2'}),
                          (1, {'name': 'BB'})],
                          {0: 1, 1: 2},
                          {0: 6, 1: 9}
                        ),
                        # third and first residue name-based
                        ([(0, {'name': 'BB'}),
                          (1, {'name': 'BB1'})],
                          {0: 0, 1: 2},
                          {0: 0, 1: 10}
                        ),
                        # selection type based
                        ([(0, {'name': 'BB', 'atype': 'A'}),
                          (1, {'atype': 'B'})],
                          {0: 0, 1: 2},
                          {0: 0, 1: 10}
                        ),
                        ))
def test_match_link_and_residue_atoms(example_meta_molecule,
                                      link_nodes,
                                      link_to_resid,
                                      expected):

    link = nx.Graph()
    link.add_nodes_from(link_nodes)
    match = polyply.src.apply_links.match_link_and_residue_atoms(example_meta_molecule,
                                                                 link,
                                                                 link_to_resid)
    assert match == expected

@pytest.mark.parametrize('link_nodes, link_to_resid',(
                        # one non-matching name
                        ([(0, {'name': 'QB1'}),
                          (1, {'name': 'BB'})],
                          {0: 0, 1: 1}
                        ),
                        # both names not matching
                        ([(0, {'name': 'QB2'}),
                          (1, {'name': 'QB'})],
                          {0: 1, 1: 2}
                        ),
                        # one additional attribute not matching
                        ([(0, {'name': 'BB', 'atype': 'Q'}),
                          (1, {'name': 'BB1', 'charge': -1})],
                          {0: 0, 1: 2}
                        ),
                        # more than one matching atom
                        ([(0, {'atype': 'A'}),
                          (1, {'other': 'A'})],
                          {0: 0, 1: 2}
                        ),
                        ))
def test_match_link_and_residue_atoms_fail(example_meta_molecule,
                                      link_nodes,
                                      link_to_resid):

    link = nx.Graph()
    link.add_nodes_from(link_nodes)

    with pytest.raises(polyply.src.apply_links.MatchError):
         polyply.src.apply_links.match_link_and_residue_atoms(example_meta_molecule,
                                                              link,
                                                              link_to_resid)

@pytest.mark.parametrize('link_defs, link_to_resids, inter_type, link_inters, expected_nodes, expected_inters',
                        (
                        # simple check single residue
                        ([[(0, {'name': 'BB1'}),
                          (1, {'name': 'BB'})]],
                         [{0: 0, 1: 1}],
                         'bonds',
                         [[vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '500'],
                                                        meta={})]],
                         [(1, 4, 1)],
                         [(0, 0)]
                        ),
                        # simple check single residue reverse
                        ([[(0, {'name': 'BB'}),
                          (1, {'name': 'BB1'})]],
                         [{0: 1, 1: 0}],
                         'bonds',
                         [[vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '500'],
                                                        meta={})]],
                         [(4, 1, 1)],
                         [(0, 0)]
                        ),
                        # add replace with node
                        ([[(0, {'name': 'BB1', 'replace': {'charge': 1.0}}),
                          (1, {'name': 'BB'})]],
                         [{0: 0, 1: 1}],
                         'bonds',
                         [[vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '500'],
                                                        meta={})]],
                         [(1, 4, 1)],
                         [(0, 0)]
                        ),
                      # test multiple versions don't overwrite
                        ([[(0, {'name': 'BB1', 'replace': {'charge': 1.0}}),
                          (1, {'name': 'BB'})]],
                         [{0: 0, 1: 1}],
                         'bonds',
                         [[vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '500'],
                                                        meta={"version": 1}),
                          vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '600'],
                                                        meta={"version": 2})]],
                          [(1, 4, 1), (1, 4, 2)],
                          [(0, 0), (0, 1)]
                        ),
                      # test multiple links which don't overwrite
                        ([[(0, {'name': 'BB1', 'replace': {'charge': 1.0}}),
                           (1, {'name': 'BB'})],
                         [(0, {'name': 'BB2'}), (1, {'name': 'BB'})]],
                         [{0: 0, 1: 1},
                          {0: 1, 1: 2}],
                         'bonds',
                         [[vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '500'],
                                                        meta={"version": 1})],
                          [vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '600'],
                                                        meta={"version": 1})]],
                          [(1, 4, 1), (6, 9, 1)],
                          [(0, 0), (1, 0)]
                        ),
                      # test multiple links which do overwrite
                        ([[(0, {'name': 'BB1'}), (1, {'name': 'BB'})],
                          [(0, {'name': 'BB1'}), (1, {'name': 'BB'})]],
                         [{0: 0, 1: 1},
                          {0: 0, 1: 1}],
                         'bonds',
                         [[vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '500'],
                                                        meta={"version": 1})],
                          [vermouth.molecule.Interaction(atoms=(0, 1),
                                                        parameters=['1', '0.33', '600'],
                                                        meta={"version": 1})]],
                          [(1, 4, 1)],
                          [(1, 0)]
                        ),
                        ))
def test_apply_link_to_residue(example_meta_molecule,
                               link_defs,
                               link_to_resids,
                               inter_type,
                               link_inters,
                               expected_nodes,
                               expected_inters):
    links = []
    processor = ApplyLinks()
    for link_nodes, link_to_resid, interactions in zip(link_defs,
                                                       link_to_resids,
                                                       link_inters):
        link = Link()
        link.add_nodes_from(link_nodes)
        link.interactions[inter_type] = interactions
        processor.apply_link_between_residues(example_meta_molecule,
                                              link,
                                              link_to_resid)

    for nodes, inter_idx in zip(expected_nodes, expected_inters):
        interaction = link_inters[inter_idx[0]][inter_idx[1]]
        new_interactions = processor.applied_links[inter_type][nodes][0]
        assert new_interactions.atoms == tuple(nodes[:-1])
        assert new_interactions.parameters == interaction.parameters
        assert new_interactions.meta == interaction.meta


@pytest.mark.parametrize('links, interactions, edges, inttype',(
   ("""
   [ link ]
   [ molmeta ]
   by_atom_ID true
   [ bonds ]
   1   2  1  0.350  1250
   """,
   [vermouth.molecule.Interaction(atoms=(0, 1),
                                  parameters=['1', '0.350', '1250'],
                                  meta={})],
   1,
   'bonds'
   ),
   ("""
   [ link ]
   [ molmeta ]
   by_atom_ID true
   [ angles ]
   2  3  4  1  125  250
   """,
   [vermouth.molecule.Interaction(atoms=(1, 2, 3),
                                  parameters=['1', '125', '250'],
                                  meta={})],
   2,
   'angles')
   ))
def test_add_explicit_link(links, interactions, edges, inttype):
    lines = """
    [ moleculetype ]
    GLY  1
    [ atoms ]
    ;id  type resnr residu atom cgnr   charge
     1   SP2   1     GLY    BB     1      0
    [ moleculetype ]
    ALA  1
    [ atoms ]
    ;id  type resnr residu atom cgnr   charge
     1   SP2   1     ALA    BB     1      0
     2   SC2   1     ALA    SC1     1      0
    """
    lines = lines + links
    lines = textwrap.dedent(lines).splitlines()
    force_field = vermouth.forcefield.ForceField(name='test_ff')
    vermouth.ffinput.read_ff(lines, force_field)

    new_mol = force_field.blocks['GLY'].to_molecule()
    new_mol.merge_molecule(force_field.blocks['GLY'])
    new_mol.merge_molecule(force_field.blocks['ALA'])

    polyply.src.apply_links.apply_explicit_link(
        new_mol, force_field.links[0])
    assert new_mol.interactions[inttype] == interactions
    assert len(new_mol.edges) == edges

@pytest.mark.parametrize('links, error_type',
                         (("""
     [ link ]
     [ molmeta ]
     by_atom_ID true
     [ bonds ]
     BB  +BB  1  0.350  1250
     """,
                           ValueError
                           ),
                          ("""
     [ link ]
     [ molmeta ]
     by_atom_ID true
     [ angles ]
     1   8  1  125  250
     """,
                           IOError)
                          ))
def test_explicit_link_failure(links, error_type):
    lines = """
    [ moleculetype ]
    GLY  1
    [ atoms ]
    ;id  type resnr residu atom cgnr   charge
     1   SP2   1     GLY    BB     1      0
    [ moleculetype ]
    ALA  1
    [ atoms ]
    ;id  type resnr residu atom cgnr   charge
     1   SP2   1     ALA    BB     1      0
     2   SC2   1     ALA    SC1    1      0
    """
    lines = lines + links
    lines = textwrap.dedent(lines).splitlines()
    force_field = vermouth.forcefield.ForceField(name='test_ff')
    vermouth.ffinput.read_ff(lines, force_field)

    new_mol = force_field.blocks['GLY'].to_molecule()
    new_mol.merge_molecule(force_field.blocks['GLY'])
    new_mol.merge_molecule(force_field.blocks['ALA'])

    with pytest.raises(error_type):
        polyply.src.apply_links.apply_explicit_link(
            new_mol, force_field.links[0])

def test_expand_exclusions():
    mol = vermouth.Molecule()
    mol.nrexcl = 1
    mol.add_edges_from([(0, 1), (1, 2), (2, 3), (2, 4)])
    nx.set_node_attributes(mol, {0:1, 1:2, 2:2, 3:2, 4:2}, "exclude")
    mol = polyply.src.apply_links.expand_excl(mol)
    ref_excl = [frozenset([0, 1]),
                frozenset([1, 2]),
                frozenset([2, 0]),
                frozenset([2, 3]),
                frozenset([2, 4]),
                frozenset([3, 4]),
                frozenset([3, 1]),
                frozenset([4, 1])]

    assert len(ref_excl) == len(mol.interactions["exclusions"])
    for excl in mol.interactions["exclusions"]:
        assert frozenset(excl.atoms) in ref_excl



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

from collections import defaultdict
import textwrap
import pytest
import vermouth.forcefield
import vermouth.molecule
import polyply
from polyply.src.topology import Topology


class TestTopParsing:
    @staticmethod
    @pytest.mark.parametrize('lines, attr, value', (
        # Refers, by index, to an atom out of range
        ("""
        [ defaults ]
        1.0   1.0   no   1.0     1.0
        """,
         "defaults",
         {"nbfunc": 1.0,
          "comb-rule": 1.0,
          "gen-pairs": "no",
          "fudgeLJ": 1.0,
          "fudgeQQ": 1.0}
         ),
        #martini style defaults
        #gen-pairs set by default
        ("""
        [ defaults ]
        1.0   1.0
        """,
         "defaults",
         {"nbfunc": 1.0,
          "comb-rule": 1.0,
          "gen-pairs": "no"}
         ),
        # martini like atomtypes
        ("""
        [ atomtypes ]
        P6 72.0 0.000 A 0.0 0.0
        """,
         "atom_types",
         {"P6": {"mass": 72.0,
                "atom_num": None,
                "charge": 0.000,
                "ptype": "A",
                "bond_type": None,
                "nb1": 0.0,
                "nb2": 0.0}}
         ),
        # amber/charrm like atomtypes
        ("""
        [ atomtypes ]
        H0           1       1.008   0.0000  A   2.47135e-01  6.56888e-02
        """,
        "atom_types",
        {"H0": {"mass": 1.008,
                "atom_num": 1.0,
                "charge": 0.000,
                "ptype": "A",
                "bond_type": None,
                "nb1": 2.47135e-01,
                "nb2": 6.56888e-02}}
         ),
        # check OPLS type atom defs.
        ("""
        [ atomtypes ]
        opls_001   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        """,
         "atom_types",
        {"opls_001":  {"mass": 12.01100,
                       "atom_num": 6,
                       "charge": 0.500,
                       "ptype": "A",
                       "bond_type": "C",
                       "nb1": 3.75000e-01,
                       "nb2": 4.39320e-01}}
         ),
        ("""
        [ nonbond_params ]
        OM      O         1    1.9670816e-03  8.5679450e-07
        """,
        "nonbond_params",
        {frozenset(["OM", "O"]):{"f": 1,
                                 "nb1": 1.9670816e-03,
                                 "nb2": 8.5679450e-07}}
        ),
        ("""
        [ bondtypes ]
        OM      O         1    1.9670816e-03  8.5679450e-07
        """,
        "types",
        {"bonds": {("OM", "O"): (["1", "1.9670816e-03", "8.5679450e-07"], None)}}
         ),
        ("""
        [ system ]
        some multiline
        description of a system
        """,
         "description",
         ["some multiline", "description of a system"]
         )))
    def test_directives(lines, attr, value):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        assert getattr(top, attr) == value

    @staticmethod
    @pytest.mark.parametrize('def_statements, bonds',
         (("""#ifdef FLEXIBLE
         [ bonds ]
         1  2
         2  3
         #endif""",
         [vermouth.molecule.Interaction(
             atoms=[0, 1], parameters=[], meta={"ifdef": "FLEXIBLE"}),
          vermouth.molecule.Interaction(
              atoms=[1, 2], parameters=[], meta={"ifdef": "FLEXIBLE"})]),
         ("""#ifndef FLEXIBLE
         [ bonds ]
         1   2
         2   3
         #endif""",
         [vermouth.molecule.Interaction(
             atoms=[0, 1], parameters=[], meta={"ifndef": "FLEXIBLE"}),
          vermouth.molecule.Interaction(
             atoms=[1, 2], parameters=[], meta={"ifndef": "FLEXIBLE"})]),
         ("""[ bonds ]
         1   2
         #ifdef FLEXIBLE
         2   3
         #endif
         3  4""",
         [vermouth.molecule.Interaction(
             atoms=[0, 1], parameters=[], meta={}),
          vermouth.molecule.Interaction(
             atoms=[1, 2], parameters=[], meta={"ifdef": "FLEXIBLE"}),
          vermouth.molecule.Interaction(
             atoms=[2, 3], parameters=[], meta={})])
         ))
    def test_ifdefs(def_statements, bonds):
        """
        test the handling if ifdefs and ifndefs at
        different positions in itp-file
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        """
        new_lines = lines + def_statements
        new_lines = textwrap.dedent(new_lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        print(top.force_field.blocks["GLY"].interactions['bonds'])
        assert top.force_field.blocks["GLY"].interactions['bonds'] == bonds

    @staticmethod
    @pytest.mark.parametrize('def_statements',
          ("""
          #ifdef random
          #include "random.itp"
          #endif
          """,
          """
          #define random
          #ifndef random
          #include "random.itp"
          #endif
          """))
    def test_ifdefs_files(def_statements):
        """
        test files in ifdef/ifndef statements are only
        read if a define has been read before
        """
        new_lines = textwrap.dedent(def_statements)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        assert True

    @staticmethod
    @pytest.mark.parametrize('def_statements, result',
        (("""#define pH5""",
          {"pH5": True}
         ),
        ("""#define ga25 1.025
         """,
          {"ga25": ["1.025"]}
        ),
        ("""
         [ moleculetype ]
         GLY 1
         [ atoms ]
         1 P4 1 ALA BB 1
         #define FLEXIBLE
         """,
        {"FLEXIBLE": True}
        )))
    def test_define(def_statements, result):
        """
        test the handling of define statements in
        """
        new_lines = textwrap.dedent(def_statements)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        assert top.defines == result

    @staticmethod
    def test_itp_handling():
       """
       test if multiple itps are read correctly
       """
       lines = """
       [ defaults ]
       1   1   no   1.0     1.0
       [ moleculetype ]
       GLY 1
       [ atoms ]
       1 P4 1 ALA BB 1
       [ moleculetype ]
       LYS 1
       [ atoms ]
       1 P4 1 ALA BB 1
       [ molecules ]
       LYS  2
       """
       new_lines = textwrap.dedent(lines).splitlines()
       force_field = vermouth.forcefield.ForceField(name='test_ff')
       top = Topology(force_field, name="test")
       polyply.src.top_parser.read_topology(new_lines, top)
       assert top.defaults == {"nbfunc": 1,
                               "comb-rule": 1,
                               "gen-pairs": "no",
                               "fudgeLJ": 1.0,
                               "fudgeQQ": 1.0}
       # check that there are two LYS in topology
       assert len(top.molecules) == 2
       for mol in top.molecules:
           assert mol.mol_name == "LYS"
       # check that both are not the same instance of meta_molecule
       assert top.molecules[0] is not top.molecules[1]
       # also check that the high-res molecule is not the same
       assert top.molecules[0].molecule is not top.molecules[1].molecule
       # we also should have each molecule as block in the force
       # field. Again we only check that it is there because
       # the actual parsing is done elsewhere
       assert len(force_field.blocks) == 2

    @staticmethod
    @pytest.mark.parametrize('lines', (
        """
         #fdef FLEXIBLE
         #define A
         #endif
         """,
        """
         #ifdef FLEXIBLE
         """,
        """
         #endif
         """,
        """
         #ifdef FLEXIBLE
         #define A
         #ifdef RANDOM
         #define B
         """,
        """
         #if
         """))
    def test_def_fails(lines):
        """
        test if incorrectly formatted ifdefs raise
        appropiate error
        """
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        with pytest.raises(IOError):
            polyply.src.top_parser.read_topology(new_lines, top)

    @staticmethod
    def test_buckingham_fail():
        lines = """
        [ defaults ]
        2   1   no   1.0     1.0
        """
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        with pytest.raises(IOError):
            polyply.src.top_parser.read_topology(new_lines, top)

    @staticmethod
    def test_atom_type_fail():
        lines = """
        [ atomtypes ]
        opls_001   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 random; SIG
        """
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        with pytest.raises(IOError):
            polyply.src.top_parser.read_topology(new_lines, top)

    @staticmethod
    @pytest.mark.parametrize('lines', (
        """
        ********
        [ defaults ]
        1   1   no   1.0     1.0
        """,
        """
        [ defaults ]
        1   1   no   1.0     1.0
        [ cmaptypes ]
        something something
        """,
        """
        [ defaults ]
        1   1   no   1.0     1.0
        [implicit_genborn_params]
        something something
        """
         ))
    def test_skip_directives(lines):
        """
        Test that directives which currently cannot be read and
        or interpreted are simply skipped and don't cause an
        error when reading the topology file.
        """
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top = Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        assert top.defaults == {"nbfunc":1.0,
                                "comb-rule":1.0,
                                "gen-pairs":'no',
                                "fudgeLJ":1.0,
                                "fudgeQQ":1.0}



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing
import itertools
import numpy as np
import scipy.optimize
import networkx as nx
from tqdm import tqdm
from polyply import jit
from .processor import Processor
from .generate_templates import find_atoms
from .linalg_functions import rotate_xyz
from .graph_utils import find_connecting_edges
"""
Processor implementing a template based back
mapping to lower resolution coordinates for
a meta molecule.
"""
def _norm_matrix(matrix):
    norm = np.sum(matrix * matrix)
    return norm
norm_matrix = jit(_norm_matrix)

def orient_template(meta_molecule, current_node, template, built_nodes):
    """
    Given a `template` and a `node` of a `meta_molecule` at lower resolution
    find the bonded interactions connecting the higher resolution template
    to its neighbours and orient a template such that the atoms point torwards
    the neighbours. In case some nodes of meta_molecule have already been built
    at the lower resolution they can be provided as `built_nodes`. In case the
    lower resolution is already built the atoms will be oriented torward the lower
    resolution atom participating in the bonded interaction.

    Parameters:
    -----------
    meta_molecule: :class:`polyply.src.meta_molecule`
    current_node:
        node key of the node in meta_molecule to which template referes to
    template: dict[collections.abc.Hashable, np.ndarray]
        dict of positions referring to the lower resolution atoms of node
    built_nodes: list
        list of meta_molecule node keys of residues that are already built

    Returns:
    --------
    dict
        the oriented template
    """
    # 1. find neighbours at meta_mol level
    neighbours = nx.all_neighbors(meta_molecule, current_node)
    current_resid = meta_molecule.nodes[current_node]["resid"]

    # 2. find connecting atoms at low-res level
    edges = []
    ref_nodes = []
    for node in neighbours:
        resid = meta_molecule.nodes[node]["resid"]
        edge = find_connecting_edges(meta_molecule,
                                     meta_molecule.molecule,
                                     (node, current_node))
        edges += edge
        ref_nodes.extend([node]*len(edge))

    # 3. build coordinate system
    ref_coords = np.zeros((3, len(edges)))
    opt_coords = np.zeros((3, len(edges)))

    for ndx, edge in enumerate(edges):
        for atom in edge:
            resid = meta_molecule.molecule.nodes[atom]["resid"]
            if resid == current_resid:
                current_atom = atom
            else:
                ref_atom = atom
                ref_resid = resid

        # the reference residue has already been build so we take the lower
        # resolution coordinates as reference
        if ref_resid in built_nodes:
            atom_name = meta_molecule.molecule.nodes[current_atom]["atomname"]

            # record the coordinates of the atom that is rotated
            opt_coords[:, ndx] = template[atom_name]

            # given the reference atom that already exits translate it to the origin
            # of the rotation, this will be the reference point for rotation
            ref_coords[:, ndx] = meta_molecule.molecule.nodes[ref_atom]["position"] -\
                                 meta_molecule.nodes[current_node]["position"]

        # the reference residue has not been build the CG center is taken as
        # reference
        else:
            atom_name = meta_molecule.molecule.nodes[current_atom]["atomname"]
            cg_node = ref_nodes[ndx] #find_atoms(meta_molecule, "resid", ref_resid)[0]

            # record the coordinates of the atom that is rotated
            opt_coords[:, ndx] = template[atom_name]

            # as the reference atom is not built take the cg node as reference point
            # for rotation; translate it to origin
            ref_coords[:, ndx] = meta_molecule.nodes[cg_node]["position"] -\
                                 meta_molecule.nodes[current_node]["position"]


    # 4. optimize the distance between reference nodes and nodes to be placed
    # only using rotation of the complete template
    #@profile
    def target_function(angles):
        rotated = rotate_xyz(opt_coords, angles[0], angles[1], angles[2])
        diff = rotated - ref_coords
        score = norm_matrix(diff)
        return score

    # choose random starting angles
    angles = np.random.uniform(low=0, high=2*np.pi, size=(3))
    opt_results = scipy.optimize.minimize(target_function, angles, method='L-BFGS-B',
                                          options={'ftol':0.01, 'maxiter': 400})

    # 5. write the template as array and rotate it corrsponding to the result above
    template_arr = np.zeros((3, len(template)))
    key_to_ndx = {}
    for ndx, key in enumerate(template.keys()):
        template_arr[:, ndx] = template[key]
        key_to_ndx[key] = ndx

    angles = opt_results['x']
    template_rotated_arr = rotate_xyz(template_arr, angles[0], angles[1], angles[2])

    # 6. write the template back as dictionary
    template_rotated = {}
    for key, ndx in key_to_ndx.items():
        template_rotated[key] = template_rotated_arr[:, ndx]

    return template_rotated

class Backmap(Processor):
    """
    This processor takes a a class:`polyply.src.MetaMolecule` and
    places coordinates form a higher resolution createing positions
    for the lower resolution molecule associated with the MetaMolecule.
    """

    def __init__(self, fudge_coords=0.4, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fudge_coords = fudge_coords

    def _place_init_coords(self, meta_molecule):
        """
        For each residue in a class:`polyply.src.MetaMolecule` the
        positions of the atoms associated with that residue stored in
        attr:`polyply.src.MetaMolecule.molecule` are created from a
        template residue located in attr:`polyply.src.MetaMolecule.templates`.

        Parameters
        ----------
        meta_molecule: :class:`polyply.src.MetaMolecule`
        """
        built_nodes = []
        for node in meta_molecule.nodes:
            if  meta_molecule.nodes[node]["build"]:
                resname = meta_molecule.nodes[node]["resname"]
                cg_coord = meta_molecule.nodes[node]["position"]
                resid = meta_molecule.nodes[node]["resid"]
                high_res_atoms = meta_molecule.nodes[node]["graph"].nodes
                template = orient_template(meta_molecule, node,
                                           meta_molecule.templates[resname],
                                           built_nodes)

                for atom_high  in high_res_atoms:
                    atomname = meta_molecule.molecule.nodes[atom_high]["atomname"]
                    vector = template[atomname]
                    new_coords = cg_coord + vector * self.fudge_coords
                    meta_molecule.molecule.nodes[atom_high]["position"] = new_coords
                built_nodes.append(resid)

    def run_molecule(self, meta_molecule):
        """
        Apply placing of coordinates to meta_molecule. For more
        detail see `self._place_init_coords`.
        """
        self._place_init_coords(meta_molecule)
        return meta_molecule



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import textwrap
from vermouth.forcefield import ForceField
import polyply
from polyply.src.topology import Topology
from polyply.src.top_parser import read_topology
from .example_fixtures import example_meta_molecule

@pytest.fixture
def example_topology():
    top_lines ="""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 45.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0  1   GLY    BB   1 0.00     45
    2    N0  1   GLY    SC1  1 0.00     45
    3    N0  1   GLY    SC2  1 0.00     45
    4    N0  2   GLU    BB   2 0.00     45
    5    N0  2   GLU    SC1  2 0.00     45
    6    N0  2   GLU    SC2  2 0.00     45
    [ bonds ]
    1    2    1  0.47 2000
    2    3    1  0.47 2000
    1    4    1  0.47 2000
    4    5    1  0.47 2000
    4    6    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0  1   ASP    BB   1 0.00
    2    N0  1   ASP    SC1  1 0.00
    3    N0  1   ASP    SC2  1 0.00
    [ bonds ]
    1    2    1  0.47 2000
    2    3    1  0.47 2000
    [ system ]
    test system
    [ molecules ]
    testA 2
    testB 1
    """
    lines = textwrap.dedent(top_lines)
    lines = lines.splitlines()
    force_field = ForceField("test")
    topology = Topology(force_field)
    read_topology(lines=lines, topology=topology, cwdir="./")
    topology.preprocess()
    topology.volumes = {"GLY": 0.53, "GLU": 0.67, "ASP": 0.43}
    return topology

@pytest.mark.parametrize('start_nodes, expected', (
   # no starting nodes defined
   ([],
   {0:None, 1:None, 2:None}
   ),
   # starting node for each molecule
   (
   ["testA-GLU", "testB-ASP"],
   {0:1, 1:1, 2:0}
   ),
   # starting node different for first two molecules
   (
   ["testA#0-GLU", "testB#1-GLY" ,"testB-ASP"],
   {0:1, 1:0, 2:0}
   )))
def test_find_starting_node_from_spec(example_topology, start_nodes, expected):
   """<mol_name>#<mol_idx>-<resname>#<resid>"""
   result = polyply.src.gen_coords.find_starting_node_from_spec(example_topology,
                                                                start_nodes)
   assert result == expected

def test_check_molecule(example_meta_molecule):
    # test if check passes for connected molecule
    polyply.src.gen_coords._check_molecules([example_meta_molecule])
    # check that if we delte an edge an error is raised
    example_meta_molecule.remove_edge(0, 1)
    with pytest.raises(IOError):
        polyply.src.gen_coords._check_molecules([example_meta_molecule])



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter, defaultdict
import numpy as np
import networkx as nx
import vermouth
import vermouth.gmx
from vermouth.parser_utils import SectionLineParser
from vermouth.gmx.itp_read import ITPDirector

class PolyplyParser(ITPDirector):
    '''
    Parser for polyply input format.
    '''
    def __init__(self, force_field):
        super().__init__(force_field)
        self.citations = set()

    @SectionLineParser.section_parser('moleculetype', 'citation')
    def _parse_citation(self, line, lineno=0):
        cite_keys = line.split()
        self.current_block.citations.update(cite_keys)

    @SectionLineParser.section_parser('citations')
    def _pase_ff_citations(self, line, lineno=0):
        # parses force-field wide citations
        cite_keys = line.split()
        self.citations.update(cite_keys)

    # overwritten to allow for dangling bonds
    def _treat_block_interaction_atoms(self, atoms, context, section):
        all_references = []
        for atom in atoms:
            reference = atom[0]
            if reference.isdigit():
                if int(reference) < 1:
                    msg = ('In section {} is a negative atom reference, which is not allowed.')
                    raise IOError(msg.format(section.name))

               # The indices in the file are 1-based
                reference = int(reference) - 1
                atom[0] = reference
            else:
                msg = ('Atom names in blocks cannot be prefixed with + or -. '
                       'The name "{}", used in section "{}" of the block "{}" '
                       'is not valid in a block.')
                raise IOError(msg.format(reference, section, context.name))
            all_references.append(reference)
        return all_references

    def treat_link_multiple(self):
        """
        Iterates over a :class:`vermouth.force_field.ForceField` and
        adds version tags for all interactions within a
        :class:`vermouth.molecule.Link` that are applied to the same atoms.
        """
        for link in self.force_field.links:
            for key in link.interactions:
                terms = link.interactions[key]
                count_terms = Counter(tuple(term.atoms) for term in terms)
                for term in terms:
                    tag = count_terms[tuple(term.atoms)]
                    if tag >= 1:
                        term.meta.update({"version":tag})
                        count_terms[tuple(term.atoms)] = tag -1

    def _treat_link_atoms(self, block, link, inter_type):

        # we need to convert the atom index to an atom-name
        n_atoms = len(block.nodes)
        # the uncommented statement does not work because node and
        # atom name are couple for blocks, which is debatably useful
        #atom_names = list(nx.get_node_attributes(block, 'atomname'))
        atom_names = [block.nodes[node]["atomname"] for node in block.nodes]

        for inter_type in link.interactions:
            for interaction in link.interactions[inter_type]:
                new_atoms = []
                for atom in interaction.atoms:
                    prefix = ""
                    while atom/n_atoms >= 1:
                        atom = atom - n_atoms
                        prefix = prefix + "+"

                    new_name = prefix + atom_names[atom]
                    new_atoms.append(new_name)
                    attrs = block.nodes[atom]
                    link.add_node(new_name, **attrs)
                    order = prefix.count("+")
                    nx.set_node_attributes(link, {new_name:order}, "order")

                interaction.atoms[:] = new_atoms

        return new_atoms

    def _split_links_and_blocks(self, block):

        # Make sure to add the atomtype resdidue number etc to
        # the proper nodes.

        n_atoms = len(block.nodes)
        res_name = block.name
        prev_atoms = []
        links = []
        for key in block.interactions:
            block_interactions = []
            for interaction in block.interactions[key]:
                if any(isinstance(atom, str) for atom in interaction.atoms):
                   return

                if np.sum(np.array(interaction.atoms) > n_atoms - 1) > 0:
                   if interaction.atoms != prev_atoms:
                       prev_atoms[:] = interaction.atoms
                       new_link = vermouth.molecule.Link()
                       new_link.interactions = defaultdict(list)
                       new_link.citations = block.citations
                       new_link.name = res_name
                       links.append(new_link)
                   links[-1].interactions[key].append(interaction)
                else:
                    block_interactions.append(interaction)

            block.interactions[key] = block_interactions

        for link in links:
            self._treat_link_atoms(block, link, key)
            self.force_field.links.append(link)

    def _make_edges(self):
       for block in self.force_field.blocks.values():
           inter_types = list(block.interactions.keys())
           for inter_type in inter_types:
               block.make_edges_from_interaction_type(type_=inter_type)

       for link in self.force_field.links:
           inter_types = list(link.interactions.keys())
           for inter_type in inter_types:
               link.make_edges_from_interaction_type(type_=inter_type)

    # overwrites the finalize method to deal with dangling bonds
    # and to deal with multiple interactions in the way needed
    # for polyply to work

    def finalize(self, lineno=0):

        if self.current_meta is not None:
            raise IOError("Your #ifdef/#ifndef section is orderd incorrectly."
                          "There is no #endif for the last pragma..")

        prev_section = self.section
        self.section = []
        self.finalize_section(prev_section, prev_section)
        self.macros = {}
        self.section = None

        for block in self.force_field.blocks.values():
            block.citations.update(self.citations)
            if len(block.nodes) > 0:
                n_atoms = len(block.nodes)
                self._split_links_and_blocks(block)
                self.treat_link_multiple()
        self._make_edges()

def read_polyply(lines, force_field):
    director = PolyplyParser(force_field)
    return list(director.parse(iter(lines)))



try:
    import sys
    import random
    import json
    import numpy as np
    import networkx as nx
    from networkx.readwrite import json_graph
except ImportError:
    msg = "You are missing a library that is required to run this example test-case."
    raise ImportError(msg)

def random_linear_comb(n_mon_tot, max_mon_side, p_side_chain):
    """
    Parameters
    ----------
    n_mon_tot: int
        total number of monomers in the polymer
    max_mon_side:
        the maximum number of monomers per branched off arm
    p_side_chain: list[float]
        probabilities to add at a level corrsponding to the
        position in the list i.e. probability of adding a monomer
        to a side chain beyond the back-bone is list[0], adding a monomer branched
        off 2 level from the back-bone is list[1].

    Returns
    --------
    nx.Graph
    """
    G = nx.Graph()
    # add the first edge to get started
    G.add_edge(0, 1, **{"linktype": "a16"})
    # backbone-nodes
    bb_nodes = [0, 1]
    ndx = 1
    prev_node = "BB"
    while ndx < n_mon_tot:
        # then we make the choice back-bone or side-chain
        # but only if we did not add a side-chain before
        # because we build a comb which has only 1 side chain
        # per back-bone otherwise it's a tree
        if prev_node == "BB":
            choice = random.choices(["BB", "SC"], weights=[1-p_side_chain[0], p_side_chain[0]])[0]
        else:
            choice = "BB"

        # if choice is BB we append to the list of BB nodes and continue
        if choice == "BB":
            G.add_edge(bb_nodes[-1], ndx+1, **{"linktype": "a16"})
            ndx += 1
            bb_nodes.append(ndx)
            prev_node = "BB"

        # if choice is SC we start growing a arm
        else:
            # add the first node of the arm
            G.add_edge(bb_nodes[-1], ndx+1, **{"linktype": "a13"})
            ndx += 1
            prev_node = "SC"
            # an arm can have at most max_side_chains monomers beyond first level
            # we add them until we draw a stop
            for sc_idxs in range(1, max_mon_side):
                choice = random.choices(["stop", "add"], weights=[
                                        1-p_side_chain[sc_idxs], p_side_chain[sc_idxs]])[0]
                if choice == "stop":
                    break
                else:
                    G.add_edge(ndx, ndx+1, **{"linktype": "a16"})
                    ndx += 1
                prev_node = "SC"
                # if we exceed the max number of monomers we terminate
                if ndx >= n_mon_tot:
                    return G

        # if we exceed the max number of monomers we terminate
        if ndx >= n_mon_tot:
            return G


def probability_by_degree_of_polymerization(p, N):
    """
    Computes the probability for a chain of degree of polymerization
    N based on the extend of reaction p for linear condensation reactions.
    See Colby and Rubinstein Polymer Physics.
    """
    return N*p**(N-1)*(1-p)**2.


def __main__():
    # input arguments
    handle = sys.argv[1]
    nmols = sys.argv[2]


    weights = np.arange(2, 600)
    weight_distribution = probability_by_degree_of_polymerization(0.970, weights)
    # select 500 chains from the molecular weight distribution
    random_sample = np.random.choice(weights, p=weight_distribution/np.sum(weight_distribution),
                                     size=int(nmols))

    for idx, n_mon in enumerate(random_sample):
        graph = random_linear_comb(n_mon, 2, [0.05, 0.95, 0.95, 0.5])
        nx.set_node_attributes(graph, "DEX", "resname")

        resids = {node: node+1 for node in graph.nodes}
        nx.set_node_attributes(graph, resids, "resid")

        g_json = json_graph.node_link_data(graph)
        filename = handle + str(n_mon) + "_" + str(idx) + ".json"
        with open(filename, "w") as file_handle:
            json.dump(g_json, file_handle, indent=2)

__main__()



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pbr.version
#from .log_helpers import StyleAdapter, get_logger

__version__ = pbr.version.VersionInfo('polyply').release_string()

# Find the data directory once.
try:
    import pkg_resources
except ImportError:
    import os
    DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
    TEST_DATA = os.path.join(os.path.dirname(__file__), 'tests/test_data')
    del os
else:
    DATA_PATH = pkg_resources.resource_filename('polyply', 'data')
    TEST_DATA = pkg_resources.resource_filename('polyply', 'tests/test_data')
    del pkg_resources

del pbr

# import numba if available
import functools
try:
    from numba import jit
except:
    jit = lambda x: x  # See also https://docs.python.org/3/library/functools.html#functools.wraps
    print("INFO - Couldn't import numba. Install it for a speed boost.")
else:
    jit = functools.partial(jit,  nopython=True, cache=True, fastmath=True)

# This could be useful for the high level API
from .src.meta_molecule import (Monomer, MetaMolecule)
from .src.apply_links import ApplyLinks
from .src.map_to_molecule import MapToMolecule
from .src.gen_itp import gen_itp, gen_params
from .src.gen_coords import gen_coords
from .src.gen_seq import gen_seq
